import asyncio
import functools
from collections.abc import Generator
from dataclasses import dataclass

import click

from yanki.anki import Deck
from yanki.parser import DeckFilesParser, DeckSpec, NoteSpec
from yanki.video import VideoOptions


@dataclass(frozen=True)
class DeckSource:
    """Filter notes in decks."""

    files: list[click.File]
    tags_include: set[str] = frozenset()
    tags_exclude: set[str] = frozenset()

    def _include_note(self, note_spec: NoteSpec) -> bool:
        """Check if a note should be included based on tag filters."""
        tags = note_spec.config.tags
        # fmt: off
        return (
            self.tags_include.issubset(tags)
            and self.tags_exclude.isdisjoint(tags)
        )

    def filter_deck_spec(self, deck_spec: DeckSpec) -> Generator[DeckSpec]:
        """Filter notes in decks, only yielding decks that still have notes."""
        filtered = [
            note_spec
            for note_spec in deck_spec.note_specs
            if self._include_note(note_spec)
        ]

        if filtered:
            deck_spec.note_specs = filtered
            yield deck_spec

    def read_specs(self):
        """Read `DeckSpec`s from `files`."""
        parser = DeckFilesParser()
        for file in self.files:
            for deck_spec in parser.parse_file(file.name, file):
                yield from self.filter_deck_spec(deck_spec)

    def read(self, options: VideoOptions):
        """Read `Deck`s from `self.files`."""
        for spec in self.read_specs():
            yield Deck(spec, video_options=options)

    def read_sorted(self, options: VideoOptions):
        """Read `Deck`s from `self.files` and sort by title."""
        return sorted(self.read(options), key=lambda deck: deck.title())

    async def read_final_async(self, options: VideoOptions):
        """Read `FinalDeck`s from `self.files` (async)."""

        async def finalize_deck_async(collection, deck):
            collection.append(await deck.finalize_async())

        final_decks = []
        async with asyncio.TaskGroup() as group:
            for deck in self.read(options):
                group.create_task(finalize_deck_async(final_decks, deck))

        return final_decks

    def read_final(self, options: VideoOptions):
        """Read `FinalDeck`s from `self.files`."""
        return asyncio.run(self.read_final_async(options))

    def read_final_sorted(self, options: VideoOptions):
        """Read `FinalDeck`s from `self.files` and sort by title."""
        return sorted(self.read_final(options), key=lambda deck: deck.title)


def deck_parameters(func):
    """Add a `decks` argument along with tag filtering options to a command.

    Adds the following options:
    - -i/--include-tag: Only include notes that have all specified tags
    - -x/--exclude-tag: Exclude notes that have any of the specified tags

    The decorated function must take a `decks` parameter of type `DeckSource`.
    """

    @click.argument("decks", nargs=-1, type=click.File("r", encoding="utf_8"))
    @click.option(
        "-i",
        "--include-tag",
        multiple=True,
        default=[],
        metavar="TAG",
        help="Only include notes that have tag TAG. If specified multiple "
        "times, notes must have all TAGs.",
    )
    @click.option(
        "-x",
        "--exclude-tag",
        multiple=True,
        default=[],
        metavar="TAG",
        help="Exclude notes that have tag TAG. If specified multiple times, "
        "notes with any tag in TAGs will be excluded.",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Create DeckSource from the tag options
        kwargs["decks"] = DeckSource(
            files=kwargs.pop("decks"),
            tags_include=frozenset(kwargs.pop("include_tag")),
            tags_exclude=frozenset(kwargs.pop("exclude_tag")),
        )

        return func(*args, **kwargs)

    return wrapper



import html
import re

from yanki.cli.decks import DeckSource
from yanki.html_out import write_html
from yanki.video import VideoOptions


def test_two_decks(cache_path, deck_1_path, deck_2_path, output_path):
    files = [
        path.open("r", encoding="utf_8") for path in [deck_1_path, deck_2_path]
    ]

    write_html(
        output_path,
        cache_path,
        DeckSource(files=files).read_final_sorted(VideoOptions(cache_path)),
        flashcards=False,
    )

    index_html = (output_path / "index.html").read_text(encoding="utf_8")
    assert index_html.startswith("<!DOCTYPE html>\n")
    assert index_html.endswith("</html>\n")

    matches = re.findall(r'<a href="(deck_[^"]+)"', index_html)
    assert len(matches) == 2

    deck_path = html.unescape(matches[0])
    deck_html = (output_path / deck_path).read_text(encoding="utf_8")
    assert deck_html.startswith("<!DOCTYPE html>\n")
    assert deck_html.endswith("</html>\n")
    assert deck_html.count('<div class="note">') == 1

    deck_path = html.unescape(matches[1])
    deck_html = (output_path / deck_path).read_text(encoding="utf_8")
    assert deck_html.startswith("<!DOCTYPE html>\n")
    assert deck_html.endswith("</html>\n")
    assert deck_html.count('<div class="note">') == 1



import asyncio
import functools
import json
import logging
import re
import shutil
import sys
import tempfile
import traceback
from multiprocessing import cpu_count
from pathlib import Path

import click
import colorlog
import genanki
import yt_dlp

from yanki.anki import FINAL_NOTE_VARIABLES
from yanki.errors import ExpectedError
from yanki.html_out import write_html
from yanki.parser import NOTE_VARIABLES, find_invalid_format
from yanki.utils import add_trace_logging, find_errors, open_in_app
from yanki.video import BadURLError, FFmpegError, Video, VideoOptions

from .decks import deck_parameters
from .server import server_options

add_trace_logging()
LOGGER = logging.getLogger(__name__)

# Click path value types
WritableDirectoryPath = functools.partial(
    click.Path,
    exists=False,
    dir_okay=True,
    file_okay=False,
    writable=True,
    readable=True,
    executable=True,
)
WritableFilePath = functools.partial(
    click.Path,
    exists=False,
    dir_okay=False,
    file_okay=True,
    writable=True,
)

# Only used to pass debug logging status out to the exception handler.
global_log_debug = False


def main():  # noqa: C901 (complex)
    exit_code = 0
    try:
        cli.main(standalone_mode=False)
    except* click.Abort:
        sys.exit("Abort!")
    except* KeyboardInterrupt:
        sys.exit(130)
    except* click.ClickException as group:
        exit_code = 1
        for error in find_errors(group):
            error.show()
            exit_code = error.exit_code
    except* FFmpegError as group:
        exit_code = 1

        for error in find_errors(group):
            if global_log_debug:
                if error.stdout is not None:
                    sys.stderr.write("STDOUT:\n")
                    sys.stderr.buffer.write(error.stdout)
                    sys.stderr.write("\n")
                sys.stderr.write("STDERR:\n")
                sys.stderr.buffer.write(error.stderr)
                sys.stderr.write("\n")
                traceback.print_exception(error, file=sys.stderr)
            else:
                # FFmpeg errors contain a bytestring of ffmpeg’s output.
                sys.stderr.buffer.write(error.stderr)
                sys.stderr.write(f"\nError in {error.command}. See above.\n")
    except* ExpectedError as group:
        exit_code = 1
        for error in find_errors(group):
            print(error, file=sys.stderr)

    return exit_code


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option(
    "--cache",
    default=Path("~/.cache/yanki/").expanduser(),
    show_default=True,
    envvar="YANKI_CACHE",
    show_envvar=True,
    type=WritableDirectoryPath(path_type=Path),
    help="Path to cache for downloads and media files.",
)
@click.option(
    "--reprocess/--no-reprocess",
    help="Force reprocessing videos.",
)
@click.option(
    "-j",
    "--concurrency",
    default=cpu_count(),
    show_default=True,
    envvar="YANKI_CONCURRENCY",
    show_envvar=True,
    type=click.INT,
    help="Number of ffmpeg process to run at once.",
)
@click.pass_context
def cli(ctx, verbose, cache, reprocess, concurrency):
    """Build Anki decks from text files containing YouTube URLs."""
    if concurrency < 1:
        raise click.UsageError("--concurrency must be >= 1.")

    ensure_cache(cache)

    ctx.obj = VideoOptions(
        cache_path=cache,
        progress=verbose > 0,
        reprocess=reprocess,
        concurrency=concurrency,
    )

    # Configure logging
    global global_log_debug  # noqa: PLW0603 (global keyword)
    if verbose > 3:
        raise click.UsageError(
            "--verbose or -v may only be specified up to 3 times."
        )
    if verbose == 3:
        global_log_debug = True
        level = logging.TRACE
    elif verbose == 2:
        global_log_debug = True
        level = logging.DEBUG
    elif verbose == 1:
        level = logging.INFO
    else:
        level = logging.WARNING

    handler = colorlog.StreamHandler()
    handler.setFormatter(
        colorlog.ColoredFormatter(
            "%(log_color)s%(levelname)s%(reset)s "
            "%(light_cyan)s%(name)s:%(reset)s %(message)s",
            log_colors={
                "TRACE": "bold_purple",
                "DEBUG": "bold_white",
                "INFO": "bold_green",
                "WARNING": "bold_yellow",
                "ERROR": "bold_red",
                "CRITICAL": "bold_red",
            },
        )
    )

    logging.basicConfig(level=level, handlers=[handler])


@cli.command()
@deck_parameters
@click.option(
    "-o",
    "--output",
    type=WritableFilePath(),
    help="Path to save decks to. Defaults to saving indivdual decks to their "
    "own files named after their sources, but with the extension .apkg.",
)
@click.pass_obj
def build(options, decks, output):
    """Build an Anki package from deck files."""
    package = genanki.Package([])  # Only used with --output

    for deck in decks.read_final(options):
        if output is None:
            # Automatically figures out the path to save to.
            deck.save_to_file()
        else:
            deck.save_to_package(package)

    if output:
        package.write_to_file(output)
        LOGGER.info(f"Wrote decks to file {output}")


@cli.command()
@deck_parameters
@click.pass_obj
def update(options, decks):
    """Update Anki from deck files.

    This will build the .apkg file in a temporary directory that will eventually
    be deleted. It will then open the .apkg file with the `open` command.
    """
    with tempfile.NamedTemporaryFile(suffix=".apkg", delete=False) as file:
        file.close()
        package = genanki.Package([])
        for deck in decks.read_final(options):
            deck.save_to_package(package)
        LOGGER.debug(f"Wrote decks to file {file.name}")
        package.write_to_file(file.name)
        LOGGER.debug(f"Opening {file.name}")
        open_in_app([file.name])


class ListNotesCommand(click.Command):
    def format_help_text(self, ctx, formatter):
        super().format_help_text(ctx, formatter)
        formatter.write("\n  Valid format variables:\n")
        for var in sorted(FINAL_NOTE_VARIABLES):
            formatter.write(f"    * {{{var}}}\n")


@cli.command(cls=ListNotesCommand)
@deck_parameters
@click.option(
    "-f",
    "--format",
    default="{url} {clip} {direction} {text}",
    show_default=True,
    type=click.STRING,
    help="The format to output in.",
)
@click.pass_obj
def list_notes(options, decks, format):
    """List notes in deck files."""
    if find_invalid_format(format, NOTE_VARIABLES) is None:
        # Don’t need FinalNotes
        for deck in decks.read_sorted(options):
            for note in deck.notes():
                # FIXME document variables
                print(format.format(**note.variables(deck_id=deck.id())))
    else:
        if error := find_invalid_format(format, FINAL_NOTE_VARIABLES):
            sys.exit(f"Invalid variable in format: {error}")

        for deck in decks.read_final_sorted(options):
            for note in deck.notes():
                # FIXME document variables
                print(format.format(**note.variables()))


@cli.command()
@click.argument("output", type=WritableDirectoryPath(path_type=Path))
@deck_parameters
@click.option(
    "-F",
    "--flashcards/--no-flashcards",
    help="Render notes as flashcards.",
)
@click.pass_obj
def to_html(options, output, decks, flashcards):
    """Generate HTML version of decks."""
    write_html(
        output,
        options.cache_path,
        decks.read_final_sorted(options),
        flashcards=flashcards,
    )


@cli.command()
@deck_parameters
@click.option(
    "-o",
    "--output",
    type=WritableFilePath(allow_dash=True, path_type=Path),
    default="-",
    help="Path to save JSON to, or - to output to stdout.",
)
@click.option(
    "-m",
    "--copy-media-to",
    type=WritableDirectoryPath(),
    default="",
    help="Directory to copy media into (leave blank to not copy media).",
)
@click.option(
    "-p",
    "--html-media-prefix",
    default="",
    help="Prefix for media references in HTML (may be a URL).",
)
@click.pass_obj
def to_json(options, output, decks, copy_media_to, html_media_prefix):
    """Generate JSON version of decks.

    Optionally, copy the media for the decks into a directory.
    """
    decks = [
        deck.to_dict(base_url=html_media_prefix)
        for deck in decks.read_final_sorted(options)
    ]

    if copy_media_to:
        copy_media_to = Path(copy_media_to)
        copy_media_to.mkdir(parents=True, exist_ok=True)
        for deck in decks:
            for note in deck["notes"]:
                new_paths = []
                for source in note["media_paths"]:
                    destination = copy_media_to / Path(source).name
                    LOGGER.info(f"Copying media to {destination}")
                    shutil.copy2(source, destination)
                    destination.chmod(0o644)
                    new_paths.append(str(destination))

                note["media_paths"] = new_paths

    with click.open_file(output, "w", encoding="utf_8") as file:
        LOGGER.info(f"Writing JSON to {output}")
        json.dump(decks, file)


@cli.command()
@deck_parameters
@server_options
@click.option(
    "-F",
    "--flashcards/--no-flashcards",
    help="Render notes as flashcards.",
)
@click.pass_obj
def serve_http(options, decks, server, flashcards):
    """Serve HTML summary of deck on localhost:8000."""
    write_html(
        options.cache_path,
        options.cache_path,
        decks.read_final_sorted(options),
        flashcards=flashcards,
    )

    server.serve_forever(directory=options.cache_path)


@cli.command()
@click.argument("urls", nargs=-1, type=click.STRING)
@click.pass_obj
def open_videos(options, urls):
    """Download, process, and open video URLs."""
    for url in urls:
        video = Video(url, options=options)
        open_in_app([asyncio.run(video.processed_video_async())])


@cli.command()
@click.argument("files", nargs=-1, type=click.File("r", encoding="utf_8"))
@click.pass_obj
def open_videos_from_file(options, files):
    """Download videos listed in a file and open them.

    If you don’t pass any arguments this will read from stdin. Videos will be
    downloaded and minimally processed, then opened with the open command.
    """
    if len(files) == 0:
        files = [sys.stdin]

    for file in files:
        for url in _find_urls(file):
            try:
                video = Video(url, options=options)
                open_in_app([asyncio.run(video.processed_video_async())])
            except BadURLError as error:
                print(f"Error: {error}")
            except yt_dlp.utils.DownloadError:
                # yt_dlp prints the error itself.
                pass


def _find_urls(file):
    """Find URLs in a file to open.

    Ignore blank lines and # comments. URLs are separated by whitespace.
    """
    for line in file:
        line = line.strip()
        # Skip blank lines and comments
        if not line or line.startswith("#"):
            continue
        # Remove trailing comments
        line = re.split(r"\s+#", line, maxsplit=1)[0]
        for url in line.split():
            if ":" in url:
                yield url
            else:
                print(f"Does not look like a URL: {url!r}")


CACHEDIR_TAG_CONTENT = """Signature: 8a477f597d28d172789f06886806bc55
# This file is a cache directory tag created by yanki.
# For information about cache directory tags, see:
#	https://bford.info/cachedir/
#
# For information about yanki, see:
#   https://github.com/danielparks/yanki
"""


def ensure_cache(cache_path: Path):
    """Make sure cache is set up."""
    cache_path.mkdir(parents=True, exist_ok=True)

    tag_path = cache_path / "CACHEDIR.TAG"
    tag_path.write_text(CACHEDIR_TAG_CONTENT, encoding="ascii")


if __name__ == "__main__":
    # Needed to call script directly, e.g. for profiling.
    main()



import io
import logging
import os
import shutil
from pathlib import Path
from typing import Any

import pytest
from pytest_console_scripts import RunResult, ScriptRunner, _StrOrPath


@pytest.fixture(scope="session")
def bin_path(tmp_path_factory):
    """Get path to directory containing mock `open`.

    This version of `open` will just print its arguments.
    """
    path = tmp_path_factory.mktemp("bin")

    for command in ["open", "xdg-open"]:
        open_path = path / command
        open_path.write_bytes(b"#!/bin/sh\necho $*\n")
        open_path.chmod(0o755)

    return path


@pytest.fixture(scope="session")
def output_path(tmp_path_factory):
    return tmp_path_factory.mktemp("output")


@pytest.fixture(scope="session")
def cache_path(tmp_path_factory):
    return tmp_path_factory.mktemp("cache")


@pytest.fixture(scope="session")
def decks_path(tmp_path_factory):
    return tmp_path_factory.mktemp("decks")


REFERENCE_DECK_1 = """
title: Test::Reference deck
file://first.png text
"""


@pytest.fixture(scope="session")
def deck_1_path(decks_path):
    shutil.copy("test-decks/good/media/first.png", decks_path / "first.png")
    path = decks_path / "reference_1.deck"
    path.write_text(REFERENCE_DECK_1, encoding="utf_8")
    return path


REFERENCE_DECK_2 = """
title: Test::Reference deck::2
file://second.png second text
"""


@pytest.fixture(scope="session")
def deck_2_path(decks_path):
    shutil.copy("test-decks/good/media/second.png", decks_path / "second.png")
    path = decks_path / "reference_2.deck"
    path.write_text(REFERENCE_DECK_2, encoding="utf_8")
    return path


class YankiRunner(ScriptRunner):
    def __init__(
        self,
        launch_mode: str,
        rootdir: Path,
        bin_path: Path,
        cache_path: Path,
        *,
        print_result: bool = True,
    ):
        super().__init__(launch_mode, rootdir, print_result)
        self.bin_path = bin_path
        self.cache_path = cache_path

    def __repr__(self) -> str:
        return f"<YankiRunner {self.launch_mode}>"

    def run(  # noqa: PLR0913 (too many arguments)
        self,
        *arguments: _StrOrPath,
        print_result: bool | None = None,
        shell: bool = False,
        cwd: _StrOrPath | None = None,
        env: dict[str, str] | None = None,
        stdin: io.IOBase | None = None,
        check: bool = False,
        **options: Any,
    ) -> RunResult:
        old_level = logging.getLogger("yanki.parser").level
        try:
            logging.getLogger("yanki.parser").setLevel(logging.INFO)

            if env is None:
                env = {}

            # Make sure our overridden `open` is in $PATH
            if "PATH" in env:
                env["PATH"] = f"{self.bin_path}:{env['PATH']}"
            else:
                env["PATH"] = f"{self.bin_path}:{os.environ['PATH']}"

            return super().run(
                ["yanki", "--cache", self.cache_path, *arguments],
                print_result=print_result,
                shell=shell,
                cwd=cwd,
                env=env,
                stdin=stdin,
                check=check,
                **options,
            )
        finally:
            logging.getLogger("yanki.parser").setLevel(old_level)


@pytest.fixture
def yanki(
    request: pytest.FixtureRequest,
    script_cwd: Path,
    script_launch_mode: str,
    bin_path: Path,
    cache_path: Path,
) -> YankiRunner:
    return YankiRunner(
        script_launch_mode,
        script_cwd,
        bin_path,
        cache_path,
        print_result=not request.config.getoption("--hide-run-results"),
    )



from yanki.field import URL_REGEX


def find_url(input):
    return URL_REGEX.search(input)[0]


def test_easy_url():
    assert (
        find_url("foo bar https://example.com/?q=abc&t=1#fragment hey")
        == "https://example.com/?q=abc&t=1#fragment"
    )


def test_easy_url_empty_fragment():
    assert (
        find_url("foo bar https://example.com/?q=abc&t=1# hey")
        == "https://example.com/?q=abc&t=1#"
    )


def test_easy_url_uppercase():
    assert (
        find_url("foo bar HTTPS://EXAMPLE.com/?q=abc&t=1#FRAGMENT hey")
        == "HTTPS://EXAMPLE.com/?q=abc&t=1#FRAGMENT"
    )


def test_newline_url():
    assert (
        find_url("foo bar\nhttps://example.com/?q=abc&t=1#fragment\nhey")
        == "https://example.com/?q=abc&t=1#fragment"
    )


def test_bare_url():
    assert (
        find_url("https://example.com/?q=abc&t=1#fragment")
        == "https://example.com/?q=abc&t=1#fragment"
    )


def test_bare_url_empty_fragment():
    assert (
        find_url("https://example.com/?q=abc&t=1#")
        == "https://example.com/?q=abc&t=1#"
    )


def test_trailing_comma():
    assert (
        find_url("foo https://example.com/?q=abc&t=1, bar")
        == "https://example.com/?q=abc&t=1"
    )


def test_trailing_period():
    assert (
        find_url("foo https://example.com/?q=abc&t=1. bar")
        == "https://example.com/?q=abc&t=1"
    )


def test_trailing_question_mark():
    assert find_url("foo https://example.com/? bar") == "https://example.com/"


def test_trailing_colon():
    assert find_url("foo https://example.com/:\nbar") == "https://example.com/"


def test_preceeding_lt():
    assert find_url("<https://example.com/ more") == "https://example.com/"


def test_lt_gt():
    assert find_url("<https://example.com/> more") == "https://example.com/"


def test_square_brackets():
    assert find_url("[https://example.com/] more") == "https://example.com/"


def test_parentheses():
    assert find_url("(https://example.com/) more") == "https://example.com/"


# Example: https://en.wikipedia.org/wiki/Wikipedia_(disambiguation)
def test_wiki_url():
    assert find_url("http://wiki/a(b)") == "http://wiki/a(b)"


def test_parentheses_around_wiki_url():
    assert find_url("(http://wiki/a(b))") == "http://wiki/a(b)"


def test_parenthesis_before_wiki_url():
    # Not ideal, but unavoidable, I think.
    assert find_url("(http://wiki/a(b) more)") == "http://wiki/a(b"


def test_parenthesis_after_wiki_url():
    # Not ideal, but unavoidable, I think.
    assert find_url("(foo http://wiki/a(b))") == "http://wiki/a(b))"



import shutil
import sys
from collections import OrderedDict
from html import escape as h
from pathlib import Path

from yanki.utils import file_safe_name


class DeckTree:
    def __init__(self, name=None):
        self.children = OrderedDict()
        self.name = name
        self.deck = None
        self.deck_file_name = None
        self.index_file_name = None

    def __getitem__(self, key):
        if key not in self.children:
            self.children[key] = DeckTree(name=key)
        return self.children[key]

    def dig(self, path):
        if len(path):
            return self[path[0]].dig(path[1:])
        return self


def write_html(output_path, cache_path, decks, *, flashcards=False):
    """Write HTML version of decks to a path."""
    if output_path == cache_path:
        # Serving HTML from the cache; no need to copy media.
        output_media_path = None
        ensure_static_link(output_path)
    else:
        # Generating a fresh output directory; copy media.
        output_path.mkdir(parents=True, exist_ok=True)
        output_media_path = output_path
        shutil.copytree(
            path_to_web_files(), output_path / "static", dirs_exist_ok=True
        )

    if len(decks) == 1:
        # Special case: single deck goes in index.html
        deck = decks[0]
        if deck.title is None:
            sys.exit(f"Deck {deck.source_path!r} does not contain title")
        write_deck_files(
            output_path / "index.html",
            output_media_path,
            deck,
            [(name, None) for name in deck.title.split("::")],
            flashcards=flashcards,
        )
        return

    # Figure out file names for decks.
    decks_by_path = {}
    deck_tree = DeckTree()
    for deck in decks:
        if deck.title is None:
            sys.exit(f"Deck {deck.source_path!r} does not contain title")

        file_name = "deck_" + file_safe_name(deck.title) + ".html"
        html_path = output_path / file_name
        if html_path in decks_by_path:
            raise KeyError(
                f"Decks with titles {decks_by_path[html_path].title!r} and "
                f"{deck.title!r} would both to write to file {html_path!r}"
            )
        decks_by_path[html_path] = deck

        title_parts = deck.title.split("::")
        leaf = deck_tree.dig(title_parts)
        leaf.deck_file_name = file_name
        leaf.deck = deck

    write_tree_indices(
        deck_tree, output_path, output_media_path, flashcards=flashcards
    )


def write_tree_indices(
    tree, output_path, output_media_path, *, title_path=None, flashcards=False
):
    if title_path is None:
        title_path = []
    if tree.deck_file_name is None and title_path == [] and tree.name is None:
        # Anonymous root.
        if len(tree.children) == 1:
            # If there is exactly one child of the anonymous root, then skip it.
            return write_tree_indices(
                next(iter(tree.children.values())),
                output_path,
                output_media_path,
                flashcards=flashcards,
            )
        # Anonymous root has zero or more than one child deck.
        tree.name = "Decks"

    if len(tree.children) == 0:
        if tree.deck_file_name:
            # A tree with a deck is created with `deck_tree.dig()`, so it should
            # always have a name.
            assert tree.name is not None  # noqa: S101
            write_deck_files(
                output_path / tree.deck_file_name,
                output_media_path,
                tree.deck,
                [*title_path, (tree.name, tree.deck_file_name)],
                flashcards=flashcards,
            )
            return (
                f'<li><a href="{h(tree.deck_file_name)}">{h(tree.name)}'
                "</a></li>"
            )
        return ""

    # Has child decks. Must have a name.
    assert tree.name is not None  # noqa: S101
    if title_path == []:
        tree.index_file_name = "index.html"
    else:
        title_names = [name for name, _ in title_path] + [tree.name]
        tree.index_file_name = (
            "index_" + file_safe_name("::".join(title_names)) + ".html"
        )

    # This rebinds the variable to a new value instead of changing the old
    # title_path object like .append() or += would:
    title_path = [*title_path, (tree.name, tree.index_file_name)]

    list_html = [
        write_tree_indices(
            child,
            output_path,
            output_media_path,
            title_path=title_path,
            flashcards=flashcards,
        )
        for child in tree.children.values()
    ]
    list_html = "<ol>\n      " + "\n      ".join(list_html) + "\n    </ol>"
    title_html = f'<a href="{h(tree.index_file_name)}">{h(tree.name)}</a>'
    if tree.deck_file_name:
        write_deck_files(
            output_path / tree.deck_file_name,
            output_media_path,
            tree.deck,
            [*title_path, ("Deck", tree.deck_file_name)],
            flashcards=flashcards,
        )
        deck_link = f'<a href="{h(tree.deck_file_name)}">Deck</a>'
        title_html += f" ({deck_link})"
    else:
        deck_link = ""

    (output_path / tree.index_file_name).write_text(
        generate_index_html(deck_link, list_html, title_path), encoding="utf_8"
    )
    return f"""<li>
        <h3>{title_html}</h3>
        {list_html}
      </li>"""


def generate_index_html(deck_link_html, child_html, title_path):
    return f"""
        <!DOCTYPE html>
        <html>
          <head>
            <title>{title_html(title_path, add_links=False)}</title>
            <meta charset="utf-8">
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <link rel="stylesheet" href="{static_url("general.css")}">
          </head>
          <body>
            <h1>{title_html(title_path, final_link=False)}</h1>
            {deck_link_html}
            {child_html}
          </body>
        </html>
        """.replace("\n        ", "\n").lstrip()


def write_deck_files(
    html_path, output_media_path, deck, title_path, *, flashcards=False
):
    html_path.write_text(
        htmlize_deck(deck, title_path, path_prefix="", flashcards=flashcards),
        encoding="utf_8",
    )

    # Copy media to output.
    if output_media_path is not None:
        for path in deck.media_paths():
            output_path = output_media_path / Path(path).name
            shutil.copy2(path, output_path)
            # Make sure media is accessible by the web server.
            output_path.chmod(0o644)


def htmlize_deck(deck, title_path, *, path_prefix="", flashcards=False):
    if deck.title is None:
        sys.exit(f"Deck {deck.source_path!r} does not contain title")

    if flashcards:
        flashcards_html = f"""
        <link rel="stylesheet" href="{static_url("flashcards.css")}">
        <script src="{static_url("flashcards.js")}" async></script>
        """
    else:
        flashcards_html = ""

    output = f"""
    <!DOCTYPE html>
    <html>
      <head>
        <title>{title_html(title_path, add_links=False)}</title>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <link rel="stylesheet" href="{static_url("general.css")}">
        {flashcards_html}
      </head>
      <body>
        <h1>{title_html(title_path, final_link=False)}</h1>"""

    for note in deck.notes():
        if more_html := note.more_field().render_html(path_prefix):
            more_html = f'<div class="more">{more_html}</div>'
        if clip := note.spec.clip_or_trim():
            clip_html = h("@" + "-".join([str(time) for time in clip]))
        else:
            clip_html = ""

        video_url_html = h(note.spec.video_url())
        output += f"""
        <div class="note">
          <div class="cards">
            <h3>{note.text_field().render_html(path_prefix)}</h3>
            <div class="media">
              {note.media_field().render_html(path_prefix)}
            </div>
          </div>
          {more_html}
          <table class="metadata">
            <tr class="direction">
              <th>Direction:</th>
              <td>media <span>{h(note.spec.direction())}</span> text</td>
            </tr>
            <tr class="note_id">
              <th>Note ID:</th>
              <td>{h(note.note_id)}</td>
            </tr>
            <tr class="source">
              <th>Source:</th>
              <td>
                <span>{h(note.spec.source_path)}</span>
                line <span>{h(str(note.spec.line_number))}</span>
              </td>
            </tr>
            <tr class="media">
              <th>Media:</th>
              <td>
                <a href="{video_url_html}">{video_url_html}</a>
                {clip_html}
              </td>
            </tr>
          </table>
        </div>"""

    return f"""{output}
      </body>
    </html>
    """.replace("\n    ", "\n").lstrip()


def title_html(title_path, *, add_links=True, final_link=True):
    if not title_path:
        return ""
    if not add_links:
        return h(" ❯ ".join([name for name, _ in title_path]))

    html = [
        f'<a href="{h(file_name)}">{h(name)}</a>' if file_name else h(name)
        for name, file_name in title_path
    ]

    if not final_link:
        # Replace final link with plain text.
        name, _ = title_path[-1]
        html[-1] = h(name)

    return " ❯ ".join(html)


def ensure_static_link(cache_path: Path):
    web_files_path = path_to_web_files()
    static_path = cache_path / "static"

    try:
        static_path.symlink_to(web_files_path)
    except FileExistsError:
        if static_path.readlink() == web_files_path:
            # Symlink already exists
            return

    try:
        static_path.unlink()
    except OSError as e:
        sys.exit(f"Error removing {static_path} to replace with symlink: {e}")

    try:
        static_path.symlink_to(web_files_path)
    except OSError as e:
        sys.exit(f"Error symlinking {static_path} to {web_files_path}: {e}")


def path_to_web_files() -> Path:
    return Path(__file__).resolve().parent.parent / "web-files"


def static_url(path) -> str:
    mtime = (path_to_web_files() / path).stat().st_mtime
    return f"static/{path}?{mtime}"



class ExpectedError(Exception):
    """An expected error that can be reported without a traceback."""


class DeckSyntaxError(ExpectedError):
    """Syntax error in a .deck file."""

    def __init__(self, message: str, source_path: str, line_number: int):
        self.message = message
        self.source_path = source_path
        self.line_number = line_number

    def __str__(self):
        return f"Error in {self.where()}: {self.message}"

    def where(self):
        return f"{self.source_path}, line {self.line_number}"



import contextlib
import dataclasses
import functools
from dataclasses import field

from yanki.field import Field, Fragment
from yanki.utils import make_frozen

# Valid variables in note_id format. Used to validate that our code uses the
# same variables in both places they’re needed.
NOTE_VARIABLES = frozenset(
    [
        # From NoteConfig:
        "crop",
        "format",
        "more",
        "overlay_text",
        "tags",
        "slow",
        "trim",
        "audio",
        "video",
        "note_id_format",
        # From NoteSpec
        "deck_id",
        "url",
        "clip",
        "direction",
        "media",
        "text",
        "line_number",
        "source_path",
    ]
)


def find_invalid_format(format, variables):
    """Validate `format`.

    Returns `KeyError` if `format` uses anything not in `variables`.
    """
    try:
        format.format(**dict.fromkeys(variables, "value"))
    except KeyError as error:
        return error
    return None


@functools.cache
def note_config_directives():
    return {field.name for field in dataclasses.fields(NoteConfig)}


@dataclasses.dataclass()
class NoteConfig:
    crop: str = "auto"
    format: str = ""
    more: Field = field(default_factory=Field)
    overlay_text: str = ""
    tags: set[str] = field(default_factory=set)
    slow: tuple[str, str | None, float] | None = None
    trim: tuple[str, str] | None = None
    audio: str = "include"
    video: str = "include"
    note_id: str = "{deck_id} {url} {clip}"

    def set(self, name, value):
        if name in note_config_directives():
            getattr(self, f"set_{name}")(value)
        else:
            raise ValueError(f"Invalid config directive {name!r}")

    def set_crop(self, input):
        self.crop = input

    def set_format(self, input):
        self.format = input

    def set_more(self, input):
        if input.startswith("+"):
            self.more.add_fragment(Fragment(input[1:]))
        else:
            self.more = Field([Fragment(input)])

    def set_overlay_text(self, input):
        if input.startswith("+"):
            self.overlay_text += input[1:]
        else:
            self.overlay_text = input

    def set_tags(self, input):
        new_tags = input.split()
        found_bare_tag = False

        for tag in new_tags:
            if tag.startswith("+"):
                self.tags.add(tag[1:])
                new_tags = None
            elif tag.startswith("-"):
                with contextlib.suppress(KeyError):
                    self.tags.remove(tag[1:])
                new_tags = None
            else:
                # No + or - prefix, which implies we’re replacing all tags.
                # FIXME: quoting so a + or - prefix can be used easily.
                found_bare_tag = True

        if found_bare_tag:
            if new_tags is None:
                raise ValueError(
                    "Invalid mix of changing tags with setting tags: "
                    f"{input.strip()}"
                )
            self.tags = set(new_tags)

    def set_slow(self, slow_spec):
        if slow_spec.strip() == "":
            self.slow = None
            return

        parts = [p.strip() for p in slow_spec.split("*")]
        if len(parts) != 2:
            raise ValueError(f"Invalid slow without '*': {slow_spec}")

        amount = float(parts[1])
        if amount < 0.01:
            raise ValueError(f"Cannot slow by less than 0.01: {slow_spec}")

        parts = [p.strip() for p in parts[0].split("-")]
        if len(parts) != 2:
            raise ValueError(f"Invalid slow without '-': {slow_spec}")

        # FIXME validate that end > start
        start = parts[0]
        if start == "":
            start = "0"

        end = parts[1]
        if end == "":
            end = None

        self.slow = (start, end, amount)

    def set_trim(self, trim):
        if trim in {"", "none"}:
            self.trim = None
        else:
            clip = [part.strip() for part in trim.split("-")]
            if len(clip) != 2:
                raise ValueError(f"trim must be time-time (found {trim!r})")
            self.trim = (clip[0], clip[1])

    def set_audio(self, audio):
        if audio in {"include", "strip"}:
            self.audio = audio
        else:
            raise ValueError('audio must be either "include" or "strip"')

    def set_video(self, video):
        if video in {"include", "strip"}:
            self.video = video
        else:
            raise ValueError('video must be either "include" or "strip"')

    def set_note_id(self, note_id):
        if error := find_invalid_format(note_id, NOTE_VARIABLES):
            raise ValueError(f"Unknown variable in note_id format: {error}")
        self.note_id = note_id

    def frozen(self):
        data = dataclasses.asdict(self)
        data["tags"] = frozenset(data["tags"])
        return NoteConfigFrozen(**data)

    def generate_note_id(self, **kwargs):
        passed_keys = set(kwargs.keys())
        if passed_keys != NOTE_VARIABLES:
            raise KeyError(
                "Incorrect variables passed to generate_note_id()\n"
                f"  unknown: {sorted(passed_keys - NOTE_VARIABLES)}\n"
                f"  missing: {sorted(NOTE_VARIABLES - passed_keys)}\n"
            )
        return self.note_id.format(**kwargs)

    def slow_spec(self):
        if self.slow is None:
            return ""

        (start, end, amount) = self.slow
        if end is None:
            end = ""

        return f"{start}-{end}*{amount}"

    def trim_spec(self):
        if self.trim is None:
            return ""

        return "-".join(self.trim)

    def variables(self):
        return {
            "crop": self.crop,
            "format": self.format,
            "more": self.more.render_html(),  # FIXME should be spec?
            "overlay_text": self.overlay_text,
            "tags": " ".join(sorted(self.tags)),
            "slow": self.slow_spec(),
            "trim": self.trim_spec(),
            "audio": self.audio,
            "video": self.video,
            "note_id_format": self.note_id,
        }

    def to_dict(self):
        """Recursively convert to dict."""
        return {
            "crop": self.crop,
            "format": self.format,
            "more": self.more.render_html(),
            "overlay_text": self.overlay_text,
            "tags": sorted(self.tags),
            "slow": self.slow,
            "trim": self.trim,
            "audio": self.audio,
            "video": self.video,
            "note_id": self.note_id,
        }


NoteConfigFrozen = make_frozen(NoteConfig)



import dataclasses
import functools

from yanki.errors import DeckSyntaxError
from yanki.parser.config import NoteConfigFrozen


@dataclasses.dataclass(frozen=True)
class NoteSpec:
    source_path: str
    line_number: int
    source: str  # config directives are stripped from this
    config: NoteConfigFrozen

    @functools.cache
    def provisional_note_id(self, deck_id="{deck_id}"):
        return self.config.generate_note_id(**self.variables(deck_id=deck_id))

    @functools.cache
    def variables(self, deck_id="{deck_id}"):
        """Get variables related to this note, including config variables."""
        return {
            **self.config.variables(),
            "deck_id": deck_id,
            "url": self.video_url(),
            "clip": self.provisional_clip_spec(),
            "direction": self.direction(),
            "media": f"{self.video_url()} {self.provisional_clip_spec()}",
            "text": self.text(),
            "line_number": self.line_number,
            "source_path": self.source_path,
        }

    def video_url(self):
        return self._parse_video_url()[0]

    @functools.cache
    def _parse_video_url(self):
        try:
            [video_url, *rest] = self.source.rstrip().split(maxsplit=1)
        except ValueError:
            self.error("NoteSpec given empty source")

        return video_url, "".join(rest)  # rest is either [] or [str]

    def clip(self):
        return self._parse_clip()[0]

    def clip_or_trim(self):
        if self.config.trim is None:
            return self.clip()  # Might be None, but that’s fine
        if self.clip() is None:
            return self.config.trim
        self.error(
            f"Clip ({self.provisional_clip_spec()!r}) is incompatible with "
            "'trim:'."
        )
        return None

    @functools.cache
    def _parse_clip(self):
        input = self._parse_video_url()[1]
        if not input.startswith("@"):
            return None, input

        [raw_clip, *rest] = input.split(maxsplit=1)
        clip = raw_clip.removeprefix("@").split("-")

        if len(clip) == 2:
            if clip[0] == "":
                clip[0] = "0"
        elif len(clip) != 1 or clip[0] == "":
            self.error(f"Invalid clip specification {raw_clip!r}")

        return clip, "".join(rest)  # rest is either [] or [str]

    def direction(self):
        return self._parse_direction()[0]

    def text(self):
        return self._parse_direction()[1]

    @functools.cache
    def _parse_direction(self):
        input = self._parse_clip()[1]

        try:
            [direction, *rest] = input.split(maxsplit=1)
            if direction in {"->", "<-", "<->"}:
                return direction, "".join(rest)  # rest is either [] or [str]
        except ValueError:
            pass

        return "<->", input

    def provisional_clip_spec(self):
        if self.clip() is None:
            return "@0-"
        return f"@{'-'.join(self.clip())}"

    def error(self, message):
        raise DeckSyntaxError(message, self.source_path, self.line_number)

    def to_dict(self):
        """Recursively convert to dict."""
        return {
            "video_url": self.video_url(),
            "clip": self.clip(),
            "direction": self.direction(),
            "text": self.text(),
            "source_path": self.source_path,
            "line_number": self.line_number,
            "source": self.source,
            "config": self.config.to_dict(),
        }


class DeckSpec:
    def __init__(self, source_path):
        self.title = None
        self.source_path = source_path
        self.note_specs = []
        # For debugging only:
        self.config = None

    def add_note_spec(self, note_spec: NoteSpec):
        self.note_specs.append(note_spec)

    def to_dict(self):
        """Recursively convert to dict."""
        return {
            "title": self.title,
            "source_path": self.source_path,
            "note_specs": [spec.to_dict() for spec in self.note_specs],
        }



from yanki.video import url_to_id

# From the following gist, modified to all use the same ID.
# https://gist.github.com/rodrigoborgesdeoliveira/987683cfbfcc8d800192da1e73adc486
YOUTUBE_ID = "lalOy8Mbfdc"
YOUTUBE_URLS = """
http://www.youtube.com/watch?v=lalOy8Mbfdc
http://youtube.com/watch?v=lalOy8Mbfdc
http://m.youtube.com/watch?v=lalOy8Mbfdc
https://www.youtube.com/watch?v=lalOy8Mbfdc
https://youtube.com/watch?v=lalOy8Mbfdc
https://m.youtube.com/watch?v=lalOy8Mbfdc

http://www.youtube.com/watch?v=lalOy8Mbfdc&feature=em-uploademail
http://youtube.com/watch?v=lalOy8Mbfdc&feature=em-uploademail
http://m.youtube.com/watch?v=lalOy8Mbfdc&feature=em-uploademail
https://www.youtube.com/watch?v=lalOy8Mbfdc&feature=em-uploademail
https://youtube.com/watch?v=lalOy8Mbfdc&feature=em-uploademail
https://m.youtube.com/watch?v=lalOy8Mbfdc&feature=em-uploademail

http://www.youtube.com/watch?v=lalOy8Mbfdc&feature=feedrec_grec_index
http://youtube.com/watch?v=lalOy8Mbfdc&feature=feedrec_grec_index
http://m.youtube.com/watch?v=lalOy8Mbfdc&feature=feedrec_grec_index
https://www.youtube.com/watch?v=lalOy8Mbfdc&feature=feedrec_grec_index
https://youtube.com/watch?v=lalOy8Mbfdc&feature=feedrec_grec_index
https://m.youtube.com/watch?v=lalOy8Mbfdc&feature=feedrec_grec_index

http://www.youtube.com/watch?v=lalOy8Mbfdc#t=0m10s
http://youtube.com/watch?v=lalOy8Mbfdc#t=0m10s
http://m.youtube.com/watch?v=lalOy8Mbfdc#t=0m10s
https://www.youtube.com/watch?v=lalOy8Mbfdc#t=0m10s
https://youtube.com/watch?v=lalOy8Mbfdc#t=0m10s
https://m.youtube.com/watch?v=lalOy8Mbfdc#t=0m10s

http://www.youtube.com/watch?v=lalOy8Mbfdc&feature=channel
http://youtube.com/watch?v=lalOy8Mbfdc&feature=channel
http://m.youtube.com/watch?v=lalOy8Mbfdc&feature=channel
https://www.youtube.com/watch?v=lalOy8Mbfdc&feature=channel
https://youtube.com/watch?v=lalOy8Mbfdc&feature=channel
https://m.youtube.com/watch?v=lalOy8Mbfdc&feature=channel

http://www.youtube.com/watch?v=lalOy8Mbfdc&playnext_from=TL&videos=osPknwzXEas&feature=sub
http://youtube.com/watch?v=lalOy8Mbfdc&playnext_from=TL&videos=osPknwzXEas&feature=sub
http://m.youtube.com/watch?v=lalOy8Mbfdc&playnext_from=TL&videos=osPknwzXEas&feature=sub
https://www.youtube.com/watch?v=lalOy8Mbfdc&playnext_from=TL&videos=osPknwzXEas&feature=sub
https://youtube.com/watch?v=lalOy8Mbfdc&playnext_from=TL&videos=osPknwzXEas&feature=sub
https://m.youtube.com/watch?v=lalOy8Mbfdc&playnext_from=TL&videos=osPknwzXEas&feature=sub

http://www.youtube.com/watch?v=lalOy8Mbfdc&feature=youtu.be
http://youtube.com/watch?v=lalOy8Mbfdc&feature=youtu.be
http://m.youtube.com/watch?v=lalOy8Mbfdc&feature=youtu.be
https://www.youtube.com/watch?v=lalOy8Mbfdc&feature=youtu.be
https://youtube.com/watch?v=lalOy8Mbfdc&feature=youtu.be
https://m.youtube.com/watch?v=lalOy8Mbfdc&feature=youtu.be

http://www.youtube.com/watch?v=lalOy8Mbfdc&feature=youtube_gdata_player
http://youtube.com/watch?v=lalOy8Mbfdc&feature=youtube_gdata_player
http://m.youtube.com/watch?v=lalOy8Mbfdc&feature=youtube_gdata_player
https://www.youtube.com/watch?v=lalOy8Mbfdc&feature=youtube_gdata_player
https://youtube.com/watch?v=lalOy8Mbfdc&feature=youtube_gdata_player
https://m.youtube.com/watch?v=lalOy8Mbfdc&feature=youtube_gdata_player

http://www.youtube.com/watch?v=lalOy8Mbfdc&list=PLGup6kBfcU7Le5laEaCLgTKtlDcxMqGxZ&index=106&shuffle=2655
http://youtube.com/watch?v=lalOy8Mbfdc&list=PLGup6kBfcU7Le5laEaCLgTKtlDcxMqGxZ&index=106&shuffle=2655
http://m.youtube.com/watch?v=lalOy8Mbfdc&list=PLGup6kBfcU7Le5laEaCLgTKtlDcxMqGxZ&index=106&shuffle=2655
https://www.youtube.com/watch?v=lalOy8Mbfdc&list=PLGup6kBfcU7Le5laEaCLgTKtlDcxMqGxZ&index=106&shuffle=2655
https://youtube.com/watch?v=lalOy8Mbfdc&list=PLGup6kBfcU7Le5laEaCLgTKtlDcxMqGxZ&index=106&shuffle=2655
https://m.youtube.com/watch?v=lalOy8Mbfdc&list=PLGup6kBfcU7Le5laEaCLgTKtlDcxMqGxZ&index=106&shuffle=2655

http://www.youtube.com/watch?feature=player_embedded&v=lalOy8Mbfdc
http://youtube.com/watch?feature=player_embedded&v=lalOy8Mbfdc
http://m.youtube.com/watch?feature=player_embedded&v=lalOy8Mbfdc
https://www.youtube.com/watch?feature=player_embedded&v=lalOy8Mbfdc
https://youtube.com/watch?feature=player_embedded&v=lalOy8Mbfdc
https://m.youtube.com/watch?feature=player_embedded&v=lalOy8Mbfdc

http://www.youtube.com/watch?app=desktop&v=lalOy8Mbfdc
http://youtube.com/watch?app=desktop&v=lalOy8Mbfdc
http://m.youtube.com/watch?app=desktop&v=lalOy8Mbfdc
https://www.youtube.com/watch?app=desktop&v=lalOy8Mbfdc
https://youtube.com/watch?app=desktop&v=lalOy8Mbfdc
https://m.youtube.com/watch?app=desktop&v=lalOy8Mbfdc


http://www.youtube.com/watch/lalOy8Mbfdc
http://youtube.com/watch/lalOy8Mbfdc
http://m.youtube.com/watch/lalOy8Mbfdc
https://www.youtube.com/watch/lalOy8Mbfdc
https://youtube.com/watch/lalOy8Mbfdc
https://m.youtube.com/watch/lalOy8Mbfdc

http://www.youtube.com/watch/lalOy8Mbfdc?app=desktop
http://youtube.com/watch/lalOy8Mbfdc?app=desktop
http://m.youtube.com/watch/lalOy8Mbfdc?app=desktop
https://www.youtube.com/watch/lalOy8Mbfdc?app=desktop
https://youtube.com/watch/lalOy8Mbfdc?app=desktop
https://m.youtube.com/watch/lalOy8Mbfdc?app=desktop


http://www.youtube.com/v/lalOy8Mbfdc
http://youtube.com/v/lalOy8Mbfdc
http://m.youtube.com/v/lalOy8Mbfdc
https://www.youtube.com/v/lalOy8Mbfdc
https://youtube.com/v/lalOy8Mbfdc
https://m.youtube.com/v/lalOy8Mbfdc

http://www.youtube.com/v/lalOy8Mbfdc?version=3&autohide=1
http://youtube.com/v/lalOy8Mbfdc?version=3&autohide=1
http://m.youtube.com/v/lalOy8Mbfdc?version=3&autohide=1
https://www.youtube.com/v/lalOy8Mbfdc?version=3&autohide=1
https://youtube.com/v/lalOy8Mbfdc?version=3&autohide=1
https://m.youtube.com/v/lalOy8Mbfdc?version=3&autohide=1

http://www.youtube.com/v/lalOy8Mbfdc?fs=1&hl=en_US&rel=0
http://youtube.com/v/lalOy8Mbfdc?fs=1&hl=en_US&rel=0
http://m.youtube.com/v/lalOy8Mbfdc?fs=1&hl=en_US&rel=0
https://www.youtube.com/v/lalOy8Mbfdc?fs=1&amp;hl=en_US&amp;rel=0
https://www.youtube.com/v/lalOy8Mbfdc?fs=1&hl=en_US&rel=0
https://youtube.com/v/lalOy8Mbfdc?fs=1&hl=en_US&rel=0
https://m.youtube.com/v/lalOy8Mbfdc?fs=1&hl=en_US&rel=0

http://www.youtube.com/v/lalOy8Mbfdc?feature=youtube_gdata_player
http://youtube.com/v/lalOy8Mbfdc?feature=youtube_gdata_player
http://m.youtube.com/v/lalOy8Mbfdc?feature=youtube_gdata_player
https://www.youtube.com/v/lalOy8Mbfdc?feature=youtube_gdata_player
https://youtube.com/v/lalOy8Mbfdc?feature=youtube_gdata_player
https://m.youtube.com/v/lalOy8Mbfdc?feature=youtube_gdata_player


http://youtu.be/lalOy8Mbfdc
https://youtu.be/lalOy8Mbfdc

http://youtu.be/lalOy8Mbfdc?feature=youtube_gdata_player
https://youtu.be/lalOy8Mbfdc?feature=youtube_gdata_player

http://youtu.be/lalOy8Mbfdc?list=PLToa5JuFMsXTNkrLJbRlB--76IAOjRM9b
https://youtu.be/lalOy8Mbfdc?list=PLToa5JuFMsXTNkrLJbRlB--76IAOjRM9b

http://youtu.be/lalOy8Mbfdc&feature=channel
https://youtu.be/lalOy8Mbfdc&feature=channel

http://youtu.be/lalOy8Mbfdc?t=1
http://youtu.be/lalOy8Mbfdc?t=1s
https://youtu.be/lalOy8Mbfdc?t=1
https://youtu.be/lalOy8Mbfdc?t=1s

http://youtu.be/lalOy8Mbfdc?si=B_RZg_I-lLaa7UU-
https://youtu.be/lalOy8Mbfdc?si=B_RZg_I-lLaa7UU-
"""


def test_youtube_urls():
    for url in YOUTUBE_URLS.split():
        assert url_to_id(url) == "youtube=" + YOUTUBE_ID, (
            f"id {YOUTUBE_ID} not in {url!r}"
        )


def test_other_urls():
    assert (
        url_to_id("https://example.com/video.mp4")
        == r"https\=||example.com|video.mp4"
    )



import inspect
import io
import json
import os
import signal
import threading
import time
import urllib.error
from pathlib import Path
from urllib.request import urlopen

import psutil
import pytest


def local_tests():
    """Get test names in this file."""
    for name, value in globals().items():
        if inspect.isfunction(value) and name.startswith("test_"):
            yield name


def test_yanki_help(yanki):
    result = yanki.run("--help", print_result=False)
    try:
        assert result.returncode == 0
        assert result.stdout.startswith("Usage: yanki")
        assert result.stderr == ""

        split = result.stdout.split("Commands:")
        assert len(split) == 2, 'Expected exactly one "Commands:" in --help'
    except:
        # We only care about the --help output if it looks malformed
        result.print()
        raise

    # Extract the commands listed in --help
    commands = set()
    for line in split[1].split("\n"):
        parts = line.split(maxsplit=1)
        if not parts:
            continue
        commands.add("test_yanki_" + parts[0].replace("-", "_"))

    assert commands <= set(local_tests()), "Expected a test for every command"


def test_yanki_build(yanki, deck_1_path):
    result = yanki.run("build", deck_1_path)
    assert result.returncode == 0
    assert result.stdout == ""
    assert result.stderr == ""
    assert deck_1_path.with_suffix(".apkg").is_file()


# Fake `open` doesn’t work without subprocess.
@pytest.mark.script_launch_mode("subprocess")
def test_yanki_update(yanki, deck_1_path):
    result = yanki.run("update", deck_1_path)
    assert result.returncode == 0
    assert result.stdout.endswith(".apkg\n")
    assert Path(result.stdout[:-1]).is_file()
    assert result.stderr == ""


# This doesn’t work without subprocess.
@pytest.mark.script_launch_mode("subprocess")
def test_yanki_serve_http(yanki, deck_1_path):
    # Need to add result to an object to get it out of thread:
    results = []

    def run_yanki_serve_http():
        results.append(yanki.run("serve-http", deck_1_path))

    httpd = threading.Thread(target=run_yanki_serve_http)
    httpd.start()

    start = time.monotonic()
    html = None
    while time.monotonic() - start < 5:  # 5 second timeout
        time.sleep(0.1)
        try:
            html = urlopen("http://localhost:8000/").read()
            break
        except urllib.error.URLError as error:
            if not isinstance(error.reason, ConnectionRefusedError):
                # Not connection refused.
                raise

    for child in psutil.Process().children(recursive=False):
        # KLUDGE: Assume that yanki is the only subprocess.
        os.kill(child.pid, signal.SIGTERM)

    httpd.join()

    assert html.startswith(b"<!DOCTYPE html>\n")

    assert len(results) == 1
    result = results[0]
    assert result.returncode == -signal.SIGTERM
    # FIXME We don’t always get the “Starting HTTP server” on stdout
    assert "GET / HTTP" in result.stderr


def test_yanki_to_html(yanki, deck_1_path, output_path):
    result = yanki.run("to-html", output_path, deck_1_path)
    assert result.returncode == 0
    assert result.stderr == ""
    assert result.stdout == ""

    index_html = (output_path / "index.html").read_text(encoding="utf_8")
    assert index_html.startswith("<!DOCTYPE html>\n")
    assert index_html.endswith("</html>\n")
    assert "<h3>text</h3>" in index_html
    assert "<img " in index_html


def test_yanki_to_json(yanki, deck_1_path):
    result = yanki.run("to-json", deck_1_path)
    assert result.returncode == 0
    assert result.stderr == ""

    output = json.loads(result.stdout)
    assert len(output) == 1
    assert len(output[0]["notes"]) == 1


def test_yanki_list_notes(yanki, deck_1_path):
    result = yanki.run("list-notes", "-f", "{url}", deck_1_path)
    assert result.returncode == 0
    assert result.stdout == "file://first.png\n"
    assert result.stderr == ""


def test_yanki_list_final_notes(yanki, deck_1_path, cache_path):
    """Check that list-notes can list notes after being processed."""
    result = yanki.run("list-notes", "-f", "{media_paths}", deck_1_path)
    assert result.returncode == 0
    assert result.stdout.startswith(f"{cache_path}/processed_file\\=||")
    assert result.stderr == ""

    # Might as well check CACHEDIR.TAG too. See https://bford.info/cachedir/
    contents = (cache_path / "CACHEDIR.TAG").read_bytes()
    assert contents[:43] == b"Signature: 8a477f597d28d172789f06886806bc55"


# Fake `open` doesn’t work without subprocess.
@pytest.mark.script_launch_mode("subprocess")
def test_yanki_open_videos(yanki, deck_1_path, cache_path):
    result = yanki.run("open-videos", f"file://{deck_1_path.parent}/first.png")
    assert result.returncode == 0
    assert result.stdout.startswith(f"{cache_path}/processed_file\\=||")
    assert result.stderr == ""


# input and fake `open` don’t work without subprocess.
@pytest.mark.script_launch_mode("subprocess")
def test_yanki_open_videos_from_file(yanki, deck_1_path, cache_path):
    result = yanki.run(
        "open-videos-from-file",
        stdin=io.StringIO(f"file://{deck_1_path.parent}/first.png\n"),
    )
    assert result.returncode == 0
    assert result.stdout.startswith(f"{cache_path}/processed_file\\=||")
    assert result.stderr == ""



import io
import logging
import textwrap

import pytest

from yanki.errors import DeckSyntaxError
from yanki.parser import DeckFilesParser
from yanki.utils import add_trace_logging

add_trace_logging()
logging.getLogger("yanki.parser").setLevel(logging.TRACE)


def parse_maybe_deck(contents, name="-"):
    return list(DeckFilesParser().parse_file(name, io.StringIO(contents)))


def parse_deck(contents, name="-"):
    specs = parse_maybe_deck(contents, name=name)
    assert len(specs) == 1
    return specs[0]


def parse_deck_dedent(contents, name="-"):
    # Strip extra indents.
    return parse_deck(textwrap.dedent(contents), name=name)


def test_empty_deck():
    assert len(parse_maybe_deck("")) == 0


def test_group():
    deck = parse_deck_dedent(
        """
        title: a
        version: 1
        overlay_text: deck
        file:///one outer1
        group:
            overlay_text: group1
            file:///two group1a
            file:///three group1b

            overlay_text: group1b
            file:///four group1c

            group:
                overlay_text: + group2
                file:///five group2a
        file:///six outer2
        """
    )
    assert deck.config.overlay_text == "deck"
    assert len(deck.note_specs) == 6

    assert deck.note_specs[0].video_url() == "file:///one"
    assert deck.note_specs[0].text() == "outer1"
    assert deck.note_specs[0].config.overlay_text == "deck"

    assert deck.note_specs[1].video_url() == "file:///two"
    assert deck.note_specs[1].text() == "group1a"
    assert deck.note_specs[1].config.overlay_text == "group1"

    assert deck.note_specs[2].video_url() == "file:///three"
    assert deck.note_specs[2].text() == "group1b"
    assert deck.note_specs[2].config.overlay_text == "group1"

    assert deck.note_specs[3].video_url() == "file:///four"
    assert deck.note_specs[3].text() == "group1c"
    assert deck.note_specs[3].config.overlay_text == "group1b"

    assert deck.note_specs[4].video_url() == "file:///five"
    assert deck.note_specs[4].text() == "group2a"
    assert deck.note_specs[4].config.overlay_text == "group1b group2"

    assert deck.note_specs[5].video_url() == "file:///six"
    assert deck.note_specs[5].text() == "outer2"
    assert deck.note_specs[5].config.overlay_text == "deck"


@pytest.mark.parametrize(
    ("lines", "more_html"),
    [
        (["more: one"], "one"),
        (["more: one", "more: two"], "two"),
        (["more: one", "more: +two"], "onetwo"),
        (["more: one", "  two"], "one<br/>two"),
        (["more: html:one", "  two"], "one\ntwo"),
        (["more: html:one", "", "  two"], "one\n\ntwo"),
        (["more: html:one", "  more: two"], "one\nmore: two"),
    ],
)
def test_deck_more_parametrized(lines, more_html):
    deck = parse_deck("title: a\n" + "\n".join(lines))
    assert len(deck.note_specs) == 0
    assert deck.config.more.render_html() == more_html


@pytest.mark.parametrize(
    ("lines", "overlay_text"),
    [
        (["overlay_text: one"], "one"),
        (["overlay_text: one", ""], "one"),
        (["overlay_text: one", "overlay_text: two"], "two"),
        (["overlay_text: one", "overlay_text: +two"], "onetwo"),
        (["overlay_text: one", "overlay_text:"], ""),
        (["overlay_text: one", "overlay_text:     \t"], ""),
        (["overlay_text: one", "overlay_text:     \t", ""], ""),
    ],
)
def test_deck_overlay_text_parametrized(lines, overlay_text):
    deck = parse_deck("title: a\n" + "\n".join(lines))
    assert len(deck.note_specs) == 0
    assert deck.config.overlay_text == overlay_text


@pytest.mark.parametrize(
    ("lines", "overlay_text"),
    [
        (["  overlay_text:"], ""),
        (["  overlay_text:", ""], ""),
        (["  overlay_text:     \t"], ""),
        (["  overlay_text:     \t", ""], ""),
        (["  overlay_text: one"], "one"),
        (["  overlay_text: one", ""], "one"),
        (["  overlay_text: one", "    two"], "one\ntwo"),
        (["  overlay_text: one", "", "    two"], "one\n\ntwo"),
        (["  overlay_text: one", "", "      two"], "one\n\ntwo"),
        (["  overlay_text: one", "", "    two", ""], "one\n\ntwo"),
        (["  overlay_text: one", "", "    two", "    "], "one\n\ntwo"),
        (["  overlay_text: one", "", "    two", "      "], "one\n\ntwo"),
        (
            ["  overlay_text: one", "", "    two", "      three"],
            "one\n\ntwo\n  three",
        ),
        (
            ["  overlay_text: one", "", "    two", "      ", "    three"],
            "one\n\ntwo\n\nthree",
        ),
        (["  overlay_text: one", "  overlay_text: two"], "two"),
        (["  overlay_text: one", "  overlay_text: +two"], "onetwo"),
        (["  overlay_text: one", "  overlay_text:"], ""),
        (["  overlay_text: one", "  overlay_text:", ""], ""),
        (["  overlay_text: one", "  overlay_text:     \t"], ""),
        (["  overlay_text: one", "  overlay_text:     \t", ""], ""),
        # Test quotes
        (["  overlay_text: one", '    "two"'], 'one\n"two"'),
    ],
)
def test_note_overlay_text_parametrized(lines, overlay_text):
    deck = parse_deck(
        "title: a\noverlay_text: deck\nfile:///foo note\n"
        + "\n".join(lines)
        + "\n\nfile:///bar note2\n"
    )
    assert len(deck.note_specs) == 2
    assert deck.config.overlay_text == "deck"
    assert deck.note_specs[0].text() == "note"
    assert deck.note_specs[0].config.overlay_text == overlay_text
    assert deck.note_specs[1].text() == "note2"
    assert deck.note_specs[1].config.overlay_text == "deck"


@pytest.mark.parametrize(
    ("note_lines", "parsed_text"),
    [
        (["one", "  two", "  \t", "  three"], "one\ntwo\n\nthree"),
        (["one", "  two", "  ", "  three"], "one\ntwo\n\nthree"),
        (["one", "  two", " ", "  three"], "one\ntwo\n\nthree"),
        (["one", "  two", "", "  three"], "one\ntwo\n\nthree"),
        (["one", "  two", "  \tthree"], "one\ntwo\n\tthree"),
        (["one  ", ""], "one"),
        (["one  ", "  two  ", "     "], "one  \ntwo"),
        (
            ["one", "  overlay_text: ignore", "  two", "", "  three"],
            "one\ntwo\n\nthree",
        ),
        (
            ["one", "  two", "   overlay_text: not config", "  four"],
            "one\ntwo\n overlay_text: not config\nfour",
        ),
        (
            ["one", "  #two", "  three"],
            "one\n#two\nthree",
        ),
        (
            ["overlay_text: not config", "  two"],
            "overlay_text: not config\ntwo",
        ),
        (
            ["#one", "  two"],
            "#one\ntwo",
        ),
        # Test quotes
        (
            ["quotes", '  "crop: 1:1"', '  " "', "  last"],
            "quotes\ncrop: 1:1\n \nlast",
        ),
        (
            ["quotes", '  "crop: 1:1"', '  ""', "  last"],
            "quotes\ncrop: 1:1\n\nlast",
        ),
        (["quotes", '  "crop: 1:1  "', "  last"], "quotes\ncrop: 1:1  \nlast"),
        (
            ["quotes", '  "crop: 1:1"  ', "  last"],
            'quotes\n"crop: 1:1"  \nlast',
        ),
        (["quotes", '  "crop: 1:1  "'], "quotes\ncrop: 1:1"),
        (["quotes", '  "crop: 1:1"  '], 'quotes\n"crop: 1:1"'),
    ],
)
def test_note_parametrized(note_lines, parsed_text):
    deck = parse_deck(
        "title: a\nfile:///foo "
        + "\n".join(note_lines)
        + "\nfile:///bar note2\n"
    )
    assert len(deck.note_specs) == 2
    assert deck.note_specs[0].text() == parsed_text
    assert deck.note_specs[1].text() == "note2"


@pytest.mark.parametrize(
    ("deck", "message"),
    [
        ("  bad", "Error in -, line 1: Unexpected indent"),
        (
            "title: a\nfile:///foo one\n  two\n bad",
            "Error in -, line 4: Unexpected indent",
        ),
        (
            "title: a\nfile:///foo note\n  illegal2: one",
            "Error in -, line 3: Invalid config directive 'illegal2'",
        ),
        (
            "file:///foo one",
            "Error in -, line 0: Does not contain title",
        ),
        (
            "title: a\n\tmore title",
            "Error in -, line 1: Title cannot have more than one line",
        ),
        (
            "version: 2",
            "Error in -, line 1: This version of yanki only supports version 1 "
            "deck files (found '2')",
        ),
        (
            "version:",
            "Error in -, line 1: This version of yanki only supports version 1 "
            "deck files (found '')",
        ),
        (
            "title: a\ngroup:\n  title: one",
            "Error in -, line 3: Title cannot be set within group",
        ),
        (
            "title: a\ngroup:\n  version: 1",
            "Error in -, line 3: Version cannot be set within group",
        ),
        (
            "title: a\nfile:///foo note\n  title: one",
            "Error in -, line 3: Title cannot be set within note",
        ),
        (
            "title: a\nfile:///foo note\n  version: 1",
            "Error in -, line 3: Version cannot be set within note",
        ),
        (
            "title: a\nfile:///foo note\n  group:\n    foobar",
            "Error in -, line 3: Group cannot be started within note",
        ),
    ],
)
def test_errors_parametrized(deck, message):
    with pytest.raises(DeckSyntaxError) as error_info:
        parse_deck(deck)
    assert str(error_info.value) == message



from yanki.field import raw_to_html


def test_easy_url():
    assert (
        raw_to_html("a http://ex/?q=a&t=1#f b\n")
        == 'a <a href="http://ex/?q=a&amp;t=1#f">http://ex/?q=a&amp;t=1#f</a> b'
    )


def test_html():
    assert (
        raw_to_html('html: <b>"hello" &amp; & invalid OK!</a>\n')
        == ' <b>"hello" &amp; & invalid OK!</a>\n'
    )


def test_md_link():
    assert (
        raw_to_html("md:[link](http://ex/?q=a&t=1#f)\n")
        == '<p><a href="http://ex/?q=a&amp;t=1#f">link</a></p>\n'
    )


def test_rst_link():
    assert (
        raw_to_html("rst:`link <http://ex/?q=a&t=1#f>`_\n")
        == '<p><a class="reference external" href="http://ex/?q=a&amp;t=1#f">'
        "link</a></p>\n"
    )



import contextlib
import dataclasses
import inspect
import logging
import os
import shlex
import subprocess
import sys
import tempfile
import types
import typing
from functools import partial, partialmethod
from pathlib import Path
from urllib.parse import urlparse

from yanki.errors import ExpectedError


class NotFileURLError(ValueError):
    """Raised by file_url_to_path() when the parameter is not a file:// URL."""


def add_trace_logging():
    """Add `logging.TRACE` level. Idempotent."""
    try:
        logging.TRACE  # noqa: B018 (not actually useless)
    except AttributeError:
        # From user DerWeh at https://stackoverflow.com/a/55276759/1043949
        logging.TRACE = 5
        logging.addLevelName(logging.TRACE, "TRACE")
        logging.Logger.trace = partialmethod(logging.Logger.log, logging.TRACE)
        logging.trace = partial(logging.log, logging.TRACE)


@contextlib.contextmanager
def atomic_open(path: Path, *, encoding="utf_8", permissions=0o644):
    """Open a file for writing and save it atomically.

    This creates a temporary file in the same directory, writes to it, then
    replaces the target file atomically even if it already exists.
    """
    with tempfile.NamedTemporaryFile(
        mode="wb" if encoding is None else "w",
        encoding=encoding,
        dir=path.parent,
        prefix=f"working_{path.stem}",
        suffix=path.suffix,
        delete=True,
        delete_on_close=False,
    ) as temp_file:
        yield temp_file
        Path(temp_file.name).rename(path)
        # Nothing for NamedTemporaryFile to delete.
        path.chmod(permissions)


def chars_in(chars, input):
    """Return chars from `chars` that are in `input`."""
    return [char for char in chars if char in input]


def file_url_to_path(url: str) -> Path:
    """Convert a file:// URL to a Path.

    Raises NotFileURLError if the URL is not a file:// URL.
    """
    parts = urlparse(url)
    if parts.scheme.lower() != "file":
        raise NotFileURLError(url)

    # urlparse doesn’t handle file: very well:
    #
    #   >>> urlparse('file://./media/first.png')
    #   ParseResult(scheme='file', netloc='.', path='/media/first.png', ...)
    return Path(parts.netloc + parts.path)


def file_not_empty(path: Path):
    """Check that `path` is a file and is non-empty."""
    return path.exists() and path.stat().st_size > 0


def file_safe_name(name):
    """Sanitize a deck title into something safe for the file system."""
    return name.replace("/", "--").replace(" ", "_")


def find_errors(group: ExceptionGroup):
    """Get actual exceptions out of nested exception groups."""
    for error in group.exceptions:
        if isinstance(error, ExceptionGroup):
            yield from find_errors(error)
        else:
            yield error


def get_key_path(data, path: list[any]):
    """Dig into `data` following the `path` of keys.

    For example, `get_key_path(data, ["a", "b", 0]) == data["a"]["b"][0]`.
    """
    for key in path:
        data = data[key]
    return data


def make_frozen(klass):
    """Kludge to produce frozen version of dataclass."""
    name = klass.__name__ + "Frozen"
    fields = dataclasses.fields(klass)

    # This isn’t realliy necessary. It doesn’t check types. It also only handles
    # `set[...]` and not `None | set[...]`, etc.
    for f in fields:
        if typing.get_origin(f.type) is set:
            f.type = types.GenericAlias(frozenset, typing.get_args(f.type))

    namespace = {
        key: value
        for key, value in klass.__dict__.items()
        if inspect.isfunction(value)
        and key != "frozen"
        and not key.startswith("set")
        and not key.startswith("_")
    }

    return dataclasses.make_dataclass(
        name,
        fields=[(f.name, f.type, f) for f in fields],
        namespace=namespace,
        frozen=True,
    )


def open_in_app(arguments):
    """Open a file or URL in the appropriate application."""
    # FIXME only works on macOS and Linux; should handle command not found.
    if os.uname().sysname == "Darwin":
        command = "open"
    elif os.uname().sysname == "Linux":
        command = "xdg-open"
    else:
        raise ExpectedError(
            f"Don’t know how to open {arguments!r} on this platform."
        )

    command_line = [command, *arguments]
    result = subprocess.run(
        command_line,
        check=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf_8",
    )

    if result.returncode != 0:
        raise ExpectedError(
            f"Error running {shlex.join(command_line)}: {result.stdout}"
        )

    sys.stdout.write(result.stdout)



import pytest

from yanki.video import Video, VideoOptions


def get_video(tmp_path):
    cache_path = tmp_path / "cache"
    cache_path.mkdir(parents=True, exist_ok=True)

    return Video(
        "file://./test-decks/good/media/stopwatch.mp4",
        options=VideoOptions(cache_path=cache_path),
    )


def test_time_parse(tmp_path):  # noqa: PLR0915 (too many statements)
    video = get_video(tmp_path)

    assert video.get_fps() == 60

    assert video.time_to_seconds(0, on_none="a") == 0
    assert video.time_to_seconds("", on_none="a") == "a"
    assert video.time_to_seconds(None, on_none="a") == "a"

    assert round(video.time_to_seconds("123.4"), 9) == 123.4
    assert round(video.time_to_seconds("123.4s"), 9) == 123.4
    assert round(video.time_to_seconds("123.4S"), 9) == 123.4
    assert round(video.time_to_seconds("123.4ms"), 9) == 0.1234
    assert round(video.time_to_seconds("123.4MS"), 9) == 0.1234
    assert round(video.time_to_seconds("123.4Ms"), 9) == 0.1234
    assert round(video.time_to_seconds("123.4mS"), 9) == 0.1234
    assert round(video.time_to_seconds("123.4us"), 9) == 0.0001234
    assert round(video.time_to_seconds("123.4US"), 9) == 0.0001234
    assert round(video.time_to_seconds("123.4Us"), 9) == 0.0001234
    assert round(video.time_to_seconds("123.4uS"), 9) == 0.0001234
    assert round(video.time_to_seconds("-2S"), 9) == -2

    assert round(video.time_to_seconds("2 s"), 9) == 2

    with pytest.raises(ValueError) as error_info:
        video.time_to_seconds("2ks")
    assert error_info.match("could not convert string to float")

    assert round(video.time_to_seconds("0:45.6"), 9) == 45.6
    assert round(video.time_to_seconds("1:45.6"), 9) == 60 + 45.6
    assert round(video.time_to_seconds("1:1:45.6"), 9) == 3600 + 60 + 45.6
    assert (
        round(video.time_to_seconds("1:23:45.6"), 9)
        == (1 * 60 + 23) * 60 + 45.6
    )
    assert (
        round(video.time_to_seconds("02:03:05.6"), 9) == (2 * 60 + 3) * 60 + 5.6
    )

    assert round(video.time_to_seconds("-0:45.6"), 9) == -45.6
    assert round(video.time_to_seconds("-1:45.6"), 9) == -(60 + 45.6)
    assert round(video.time_to_seconds("-1:1:45.6"), 9) == -(3600 + 60 + 45.6)
    assert round(video.time_to_seconds("-1:23:45.6"), 9) == -(
        (1 * 60 + 23) * 60 + 45.6
    )
    assert round(video.time_to_seconds("-02:03:05.6"), 9) == -(
        (2 * 60 + 3) * 60 + 5.6
    )

    with pytest.raises(ValueError) as error_info:
        video.time_to_seconds(":2")
    assert error_info.match("could not convert string to float")

    with pytest.raises(ValueError) as error_info:
        video.time_to_seconds("2:")
    assert error_info.match("could not convert string to float")

    with pytest.raises(ValueError) as error_info:
        video.time_to_seconds("af:2")
    assert error_info.match("could not convert string to float")

    # FIXME? Should probably be error
    assert (
        round(video.time_to_seconds("2.1:3:5.6"), 9)
        == (2.1 * 60 + 3) * 60 + 5.6
    )

    # Frames
    assert round(video.time_to_seconds("0f"), 9) == 0
    assert round(video.time_to_seconds("1F"), 9) == round(1 / 60, 9)
    assert round(video.time_to_seconds("2F"), 9) == round(2 / 60, 9)
    assert round(video.time_to_seconds("-2F"), 9) == round(-2 / 60, 9)

    with pytest.raises(ValueError) as error_info:
        video.time_to_seconds("2fF")
    assert error_info.match("invalid literal")

    with pytest.raises(ValueError) as error_info:
        video.time_to_seconds("2F0")
    assert error_info.match("could not convert string to float")



import asyncio

from yanki.anki import FINAL_NOTE_VARIABLES, Note
from yanki.parser import NoteSpec
from yanki.parser.config import NOTE_VARIABLES, NoteConfig
from yanki.video import VideoOptions


def example_note_spec():
    return NoteSpec(
        source_path="-",
        line_number=1,
        source="file://test-decks/good/media/first.png text",
        config=NoteConfig().frozen(),
    )


def example_note(cache_path):
    return Note(example_note_spec(), VideoOptions(cache_path))


def test_note_spec_variables():
    assert set(example_note_spec().variables().keys()) == NOTE_VARIABLES


def test_note_variables(cache_path):
    assert set(example_note(cache_path).variables().keys()) == NOTE_VARIABLES


def test_final_note_variables(cache_path):
    note = asyncio.run(example_note(cache_path).finalize_async(deck_id=1))
    assert set(note.variables().keys()) == FINAL_NOTE_VARIABLES



import io
import logging
import re
from copy import deepcopy
from pathlib import Path

from yanki.errors import DeckSyntaxError
from yanki.parser.config import NoteConfig
from yanki.parser.model import DeckSpec, NoteSpec
from yanki.utils import add_trace_logging

add_trace_logging()
LOGGER = logging.getLogger(__name__)


# Regular expression to identify config directives.
#
# Indentiation must be stripped first, and `fullmatch` must be used.
CONFIG_REGEX = re.compile(
    r"""
    # Config directive must start with a letter.
    ([a-z][a-z0-9._\[\]-]*):

    # The value must be separated from the colon by whitespace, but if there is
    # no value then whitespace is not required. (Consider a file that ends with
    # a config directive and then no newline.)
    (?:\s+(\S.*\s*)?)?
    """,
    flags=re.IGNORECASE | re.VERBOSE,
)


class Parser:
    def __init__(self, line_number, deck, config):
        self.start_line_number = line_number
        self.current_line_number = line_number
        self.deck = deck
        self.config = config
        self.indent = None
        self.child_parser = None
        self.trace("Opening")

    def check_child_parser(self, indent, line):
        if self.child_parser:
            if self.child_parser.parse_line(
                self.current_line_number, indent, line
            ):
                return True
            self.child_parser = None
        return False

    def parse_line(self, line_number, indent, line):
        self.current_line_number = line_number
        self.trace(f"parse_line({indent!r}, {line!r})")

        if line.strip("\n\r") == "":
            # Blank line; indent might be wrong entirely so ignore it.
            if self.check_child_parser(indent, line):
                return True
        elif self.indent is None:
            self.indent = indent
        elif len(indent) > len(self.indent):
            if not indent.startswith(self.indent):
                self.line_error("Mismatched indent")

            if self.check_child_parser(indent, line):
                return True

            self.unexpected_indent()
            # Add extra indent back to line.
            line = indent.removeprefix(self.indent) + line
        elif len(indent) < len(self.indent):
            # Smaller indent; end of this parser. Any mismatched indent will be
            # caught in the outer parser.
            self.close()
            return False
        elif self.indent != indent:
            self.line_error("Mismatched indent")
        elif self.child_parser:
            # Indent the same, so close any child parser.
            self.child_parser.close()
            self.child_parser = None

        self.parse_unindented(line)
        return True

    def unexpected_indent(self):
        self.line_error("Unexpected indent")

    def close(self):
        if self.child_parser:
            self.child_parser.close()
            self.child_parser = None
        self.trace("Closing")

    def parse_unindented(self, line):
        if matches := CONFIG_REGEX.fullmatch(line):
            self.parse_config(matches[1], matches[2] or "")
        else:
            self.parse_text(line)

    def parse_config(self, directive, rest):
        if directive == "group":
            if rest.strip():
                self.line_error(
                    "Unexpected value after 'group:': {rest.strip()!r}"
                )
            self.child_parser = GroupParser(self)
        else:
            self.child_parser = ConfigParser(self, directive, rest)

    def parse_text(self, line):
        if line.startswith("#") or line.strip() == "":
            # Comment or blank line; ignore.
            return
        self.child_parser = NoteParser(self, line)

    def trace(self, message):
        LOGGER.trace(
            f"{self.logging_id()} at line {self.current_line_number}: {message}"
        )

    def logging_id(self):
        return self.__class__.__name__

    def parser_error(self, error):
        """Raise a syntax error located at the start of the parser."""
        raise DeckSyntaxError(
            str(error),
            self.deck.source_path,
            self.start_line_number,
        )

    def line_error(self, error):
        """Raise a syntax error located at the current line."""
        raise DeckSyntaxError(
            str(error),
            self.deck.source_path,
            self.current_line_number,
        )


class DeckParser(Parser):
    # Supports notes, config, groups, "title:", "version:"
    def __init__(self, deck):
        super().__init__(0, deck, NoteConfig())
        self.indent = ""

    def finish(self):
        """Close out inner parsers and return the finished deck."""
        self.close()
        if self.deck.title is None:
            self.parser_error("Does not contain title")
        self.deck.config = self.config
        return self.deck

    def logging_id(self):
        return f"{self.__class__.__name__}[{self.deck.source_path!r}]"


class SubParser(Parser):
    def __init__(self, parent):
        super().__init__(parent.current_line_number, parent.deck, parent.config)


class GroupParser(SubParser):
    # Supports notes, config, groups
    def __init__(self, parent):
        super().__init__(parent)
        self.config = deepcopy(self.config)

    def parse_config(self, directive, rest):
        if directive == "title":
            self.line_error("Title cannot be set within group")
        if directive == "version":
            self.line_error("Version cannot be set within group")
        super().parse_config(directive, rest)


class NoteParser(SubParser):
    # Supports text, config
    def __init__(self, parent, line):
        self.text = [line]
        super().__init__(parent)
        self.config = deepcopy(self.config)

    def unexpected_indent(self):
        pass

    def close(self):
        super().close()
        self.deck.add_note_spec(
            NoteSpec(
                config=self.config.frozen(),
                source_path=self.deck.source_path,
                line_number=self.start_line_number,
                source="".join(self.text),
            )
        )

    def parse_config(self, directive, rest):
        if directive == "title":
            self.line_error("Title cannot be set within note")
        if directive == "version":
            self.line_error("Version cannot be set within note")
        if directive == "group":
            self.line_error("Group cannot be started within note")
        super().parse_config(directive, rest)

    def parse_text(self, line):
        # Quotes can be used to prevent a line from being a config directive.
        line_chomped = line.rstrip("\n\r")
        if line.startswith('"') and line_chomped.endswith('"'):
            # Stip quotes, but add the newline back:
            line = line_chomped[1:-1] + line[len(line_chomped) :]

        self.text.append(line)

    def logging_id(self):
        return f"{self.__class__.__name__}[{self.text[0]!r}]"


class ConfigParser(SubParser):
    # Supports text
    def __init__(self, parent, directive, rest):
        self.directive = directive
        self.text = [rest]
        super().__init__(parent)

    def unexpected_indent(self):
        pass

    def parse_unindented(self, line):
        # Don’t even look for config directives.
        self.text.append(line)

    def parse_config(self, _directive, _rest):
        raise RuntimeError(
            "parse_config() should be unreachable in ConfigParser"
        )

    def close(self):
        super().close()
        value = "".join(self.text).strip()
        self.trace(f"Trying to set {value!r}")

        if self.directive == "title":
            # We prevent the ConfigParser from being created with this directive
            # in all parsers except DeckParser.
            if "\n" in value:
                self.parser_error("Title cannot have more than one line")
            self.deck.title = value
            return

        if self.directive == "version":
            # We prevent the ConfigParser from being created with this directive
            # in all parsers except DeckParser.
            if value != "1":
                self.parser_error(
                    "This version of yanki only supports version 1 deck files "
                    f"(found {value!r})"
                )
            return

        try:
            self.config.set(self.directive, value)
        except ValueError as error:
            self.parser_error(error)

    def logging_id(self):
        return f"{self.__class__.__name__}[{self.directive!r}]"


class DeckFilesParser:
    def __init__(self):
        self.finished_decks = []
        self.parser = None

    def open(self, path):
        """Open a deck file for parsing."""
        self.close()
        self.parser = DeckParser(DeckSpec(path))

    def close(self):
        """Close deck file and mark deck finished."""
        if self.parser:
            self.finished_decks.append(self.parser.finish())
        self.parser = None

    def flush_decks(self):
        finished_decks = self.finished_decks
        self.finished_decks = []
        return finished_decks

    def parse_file(self, file_name: str, file: io.TextIOBase):
        for line_number, line in enumerate(file, start=1):
            self.parse_line(file_name, line_number, line)
            yield from self.flush_decks()

        self.close()
        yield from self.flush_decks()

    def parse_path(self, path):
        with Path(path).open("r", encoding="utf_8") as file:
            yield from self.parse_file(file.name, file)

    def parse_input(self, input):
        """Parse files from FileInput."""
        for line in input:
            self.parse_line(input.filename(), input.filelineno(), line)
            yield from self.flush_decks()

        self.close()
        yield from self.flush_decks()

    def parse_line(self, path, line_number, line):
        if not self.parser or self.parser.deck.source_path != path:
            self.open(path)

        unindented = line.lstrip(" \t")
        indent = line[0 : len(line) - len(unindented)]
        if not self.parser.parse_line(line_number, indent, unindented):
            # DeckParser always has 0 indent because an indent starts a child
            # parser. If it returns False to indicate it found an outdent, then
            # something went very wrong.
            raise RuntimeError("DeckParser found impossible outdent")



from pathlib import Path

import pytest

from yanki.cli.decks import DeckSource
from yanki.utils import find_errors
from yanki.video import VideoOptions


def find_deck_files(base_path):
    yield from Path(base_path).rglob("*.deck")


def read_first_line(path: Path):
    with path.open("r", encoding="utf_8") as input:
        for line in input:
            return line
    return None


@pytest.mark.parametrize("path", find_deck_files("test-decks/errors"))
def test_deck_error(path, cache_path):
    options = VideoOptions(cache_path=cache_path)
    with (
        path.open("r", encoding="utf_8") as file,
        pytest.raises(ExceptionGroup) as error_info,
    ):
        DeckSource(files=[file]).read_final(options)

    [error] = list(find_errors(error_info.value))

    first_line = read_first_line(path)
    assert first_line[0:2] == "# "
    assert first_line[-1] == "\n"
    assert str(error) == first_line[2:-1]


@pytest.mark.parametrize("path", find_deck_files("test-decks/good"))
def test_deck_success(path, cache_path):
    options = VideoOptions(cache_path=cache_path)
    with path.open("r", encoding="utf_8") as file:
        [deck] = DeckSource(files=[file]).read_final(options)
    assert len(deck.notes()) >= 1



from pathlib import Path

import pytest

from yanki.utils import NotFileURLError, atomic_open, file_url_to_path


def test_atomic_open(tmp_path):
    path = tmp_path / "prefix.suffix"

    with atomic_open(path) as file:
        file.write("First write\n")
    assert path.read_text() == "First write\n"
    assert [path.name for path in tmp_path.iterdir()] == ["prefix.suffix"]

    with atomic_open(path) as file:
        file.write("Second write\n")
    assert path.read_text() == "Second write\n"
    assert [path.name for path in tmp_path.iterdir()] == ["prefix.suffix"]


def test_atomic_open_error(tmp_path):
    path = tmp_path / "prefix.suffix"

    with atomic_open(path) as file:
        file.write("First write\n")
    assert path.read_text() == "First write\n"
    assert [path.name for path in tmp_path.iterdir()] == ["prefix.suffix"]

    with pytest.raises(RuntimeError) as error_info:  # noqa: PT012 SIM117
        # Ignore PT012: we check for the specific `raise`.
        # Ignore SIM117: nested `with`s makes this more clear.
        with atomic_open(path) as file:
            file.write("Second write\n")
            file.close()
            raise RuntimeError("boo")
    assert error_info.match("boo")

    assert path.read_text() == "First write\n"
    assert [path.name for path in tmp_path.iterdir()] == ["prefix.suffix"]


def test_atomic_open_deleted(tmp_path):
    path = tmp_path / "prefix.suffix"

    with pytest.raises(FileNotFoundError) as error_info:  # noqa: PT012 SIM117
        # Ignore PT012: testing context manager.
        # Ignore SIM117: nested `with`s makes this more clear.
        with atomic_open(path) as file:
            Path(file.name).unlink()
            file.write("First write\n")
    assert error_info.match("No such file or directory")

    assert not path.exists()
    assert [path.name for path in tmp_path.iterdir()] == []


def test_file_url_to_path():
    with pytest.raises(NotFileURLError):
        file_url_to_path("foo")
    with pytest.raises(NotFileURLError):
        file_url_to_path("http://example.com/foo")

    assert file_url_to_path("file://a/b/c") == Path("a/b/c")

    base = Path("/BASE")
    assert base / file_url_to_path("file:///a/b/c") == Path("/a/b/c")
    assert base / file_url_to_path("file://./a/b/c") == Path("/BASE/a/b/c")
    assert base / file_url_to_path("file://") == Path("/BASE")



import functools
import http.server
import threading
import time
from dataclasses import dataclass

import click

from yanki.utils import open_in_app


@dataclass(frozen=True)
class Server:
    """Configuration for HTTP server options."""

    open: bool = False
    bind: str = "localhost:8000"

    def __post_init__(self):
        """Validate the configuration."""
        # Trigger validation by accessing the cached property
        self.bind_tuple  # noqa: B018 (not actually useless)

    @functools.cached_property
    def bind_tuple(self) -> tuple[str, int]:
        """Split --bind value into (address, port) and validate.

        Raises:
            click.UsageError: If bind format is invalid.
        """
        bind_parts = self.bind.split(":")
        if len(bind_parts) != 2:
            raise click.UsageError("--bind must be in address:port format.")

        address, port_str = bind_parts

        try:
            port = int(port_str)
        except ValueError as error:
            raise click.UsageError("--bind expects an integer port.") from error

        if not (1 <= port <= 65535):
            raise click.UsageError("--bind port must be between 1 and 65535.")

        return (address, port)

    @property
    def bind_address(self) -> str:
        """Get the address to bind to from --bind."""
        return self.bind_tuple[0]

    @property
    def bind_port(self) -> int:
        """Get the port to bind to from --bind."""
        return self.bind_tuple[1]

    def serve_forever(self, directory="."):
        """Start the HTTP server in `directory`."""

        def open_browser():
            time.sleep(0.5)
            open_in_app([f"http://localhost:{self.bind_port}/"])

        httpd = http.server.HTTPServer(
            self.bind_tuple,
            functools.partial(
                http.server.SimpleHTTPRequestHandler, directory=directory
            ),
        )

        if self.open:
            threading.Thread(target=open_browser).start()

        print(f"Starting HTTP server on http://{self.bind}/")
        httpd.serve_forever()


def server_options(func):
    """Add common server options to a command.

    Adds the following options:
    - --open: Open in web browser
    - --bind: Address to bind the HTTP server to

    The decorated function must take a `server` parameter of type `Server`.
    """

    @click.option(
        "-o",
        "--open/--no-open",
        default=False,
        help="Open the server URL in a web browser.",
    )
    @click.option(
        "-b",
        "--bind",
        default="localhost:8000",
        show_default=True,
        metavar="ADDRESS:PORT",
        help="Address and port to start the HTTP server on.",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Removes server options from kwargs:
        kwargs["server"] = Server(
            open=kwargs.pop("open"),
            bind=kwargs.pop("bind"),
        )

        return func(*args, **kwargs)

    return wrapper



import asyncio
import functools
import hashlib
import logging
from dataclasses import dataclass
from pathlib import Path

import genanki

from yanki.field import Field, Fragment, ImageFragment, VideoFragment
from yanki.parser import NOTE_VARIABLES, DeckSpec, NoteSpec
from yanki.video import Video, VideoOptions

LOGGER = logging.getLogger(__name__)


# Keep these variables local
def yanki_card_model():
    field_base = 7504631604350024486
    back_template = """
    {{FrontSide}}

    <hr id="answer">

    <p>{{$FIELD}}</p>

    {{#More}}
    <div class="more">{{More}}</div>
    {{/More}}
  """.replace("\n    ", "\n").strip()

    return genanki.Model(
        1221938102,
        "Optionally Bidirectional (yanki)",
        fields=[
            # The first field is the one displayed when browsing cards.
            {"name": "Text", "id": field_base + 1, "font": "Arial"},
            {"name": "More", "id": field_base + 4, "font": "Arial"},
            {"name": "Media", "id": field_base + 0, "font": "Arial"},
            {
                "name": "Text to media",
                "id": field_base + 2,
                "font": "Arial",
            },  # <-
            {
                "name": "Media to text",
                "id": field_base + 3,
                "font": "Arial",
            },  # ->
        ],
        templates=[
            {
                "name": "Text to media",  # <-
                "id": 6592322563225791602,
                "qfmt": "{{#Text to media}}<p>{{Text}}</p>{{/Text to media}}",
                "afmt": back_template.replace("$FIELD", "Media"),
            },
            {
                "name": "Media to text",  # ->
                "id": 6592322563225791603,
                "qfmt": "{{#Media to text}}<p>{{Media}}</p>{{/Media to text}}",
                "afmt": back_template.replace("$FIELD", "Text"),
            },
        ],
        css="""
      .card {
        font: 20px sans-serif;
        text-align: center;
        color: #000;
        background-color: #fff;
      }

      .more {
        font-size: 16px;
      }
    """,
    )


YANKI_CARD_MODEL = yanki_card_model()


def name_to_id(name):
    bytes = hashlib.sha256(name.encode("utf_8")).digest()
    # Apparently deck ID is i64
    return int.from_bytes(bytes[:8], byteorder="big", signed=True)


class Note:
    def __init__(self, spec, video_options: VideoOptions):
        self.spec = spec
        self.video_options = video_options
        self.logger = logging.getLogger(
            f"Note[{self.spec.provisional_note_id()}]"
        )

    async def finalize_async(self, deck_id):
        video = await self.video().finalize_async()
        media_path = await video.processed_video_async()

        if video.is_still() or video.output_ext() == "gif":
            media_fragment = ImageFragment(media_path, video)
        else:
            media_fragment = VideoFragment(media_path, video)

        note_id = self.note_id(deck_id)
        return FinalNote(
            note_id=note_id,
            deck_id=str(deck_id),
            media_fragment=media_fragment,
            text=self.text(),
            spec=self.spec,
            clip_spec=self.clip_spec(),
            logger=logging.getLogger(f"FinalNote[{note_id}]"),
        )

    @functools.cache
    def video(self):
        try:
            video = Video(
                self.spec.video_url(),
                working_dir=Path(self.spec.source_path).parent,
                options=self.video_options,
                logger=self.logger,
            )
            video.audio(self.spec.config.audio)
            video.video(self.spec.config.video)
            if self.spec.config.crop:
                video.crop(self.spec.config.crop)
            if self.spec.config.format:
                video.format(self.spec.config.format)
            if self.spec.config.slow:
                (start, end, amount) = self.spec.config.slow
                video.slow(start=start, end=end, amount=amount)
            if self.spec.config.overlay_text:
                video.overlay_text(self.spec.config.overlay_text)
        except ValueError as error:
            self.spec.error(error)

        clip = self.spec.clip_or_trim()
        if clip is not None:
            if len(clip) == 1:
                video.snapshot(clip[0])
            elif len(clip) == 2:
                video.clip(clip[0], clip[1])
            else:
                self.spec.error(f"Invalid clip: {clip!r}")

        return video

    # {deck_id} is just a placeholder. To get the real note_id, you need to have
    # a deck_id.
    def note_id(self, deck_id="{deck_id}"):
        return self.spec.config.generate_note_id(
            **self.variables(deck_id=deck_id),
        )

    def variables(self, deck_id="{deck_id}"):
        return {
            **self.spec.variables(),
            "deck_id": deck_id,
            "clip": self.clip_spec(),
            "media": f"{self.spec.video_url()} {self.clip_spec()}",
            "text": self.text(),
        }

    @functools.cache
    def clip_spec(self):
        if self.spec.clip() is None:
            return "@0-"
        if len(self.spec.clip()) in {1, 2}:
            return "@" + "-".join(
                [
                    str(self.video().time_to_seconds(t, on_none=""))
                    for t in self.spec.clip()
                ]
            )
        raise ValueError(f"Invalid clip: {self.spec.clip()!r}")

    def text(self):
        if self.spec.text() == "":
            return self.video().title()
        return self.spec.text()


EXTRA_FINAL_NOTE_VARIABLES = frozenset(
    [
        "note_id",
        "media_paths",
        "video_parameters",
        "auto_crop",
    ]
)

FINAL_NOTE_VARIABLES = EXTRA_FINAL_NOTE_VARIABLES | NOTE_VARIABLES

if NOTE_VARIABLES & EXTRA_FINAL_NOTE_VARIABLES:
    raise KeyError(
        "Variables in both NOTE_VARIABLES and EXTRA_FINAL_NOTE_VARIABLES: "
        + ", ".join(sorted(NOTE_VARIABLES & EXTRA_FINAL_NOTE_VARIABLES))
    )


@dataclass(frozen=True)
class FinalNote:
    deck_id: str
    note_id: str
    media_fragment: Fragment
    text: str
    spec: NoteSpec
    clip_spec: str
    logger: logging.Logger

    def media(self):
        for field in self.content_fields():
            yield from field.media()

    def media_paths(self):
        for field in self.content_fields():
            yield from field.media_paths()

    def content_fields(self):
        return [self.text_field(), self.more_field(), self.media_field()]

    def text_field(self):
        return Field([Fragment(self.text)])

    def more_field(self):
        return self.spec.config.more

    def media_field(self):
        return Field([self.media_fragment])

    def variables(self):
        return {
            **self.spec.variables(),
            "deck_id": self.deck_id,
            "note_id": self.note_id,
            "clip": self.clip_spec,
            "media": f"{self.spec.video_url()} {self.clip_spec}",
            "text": self.text,
            "media_paths": " ".join(self.media_paths()),
            "auto_crop": " / ".join(
                [str(media.cropdetect()) for media in self.media()]
            ),
            "video_parameters": " / ".join(
                [" ".join(media.parameters_list()) for media in self.media()]
            ),
        }

    def genanki_note(self):
        media_to_text = text_to_media = ""
        if self.spec.direction() == "<->":
            text_to_media = "1"
            media_to_text = "1"
        elif self.spec.direction() == "<-":
            text_to_media = "1"
        elif self.spec.direction() == "->":
            media_to_text = "1"
        else:
            raise ValueError(f"Invalid direction {self.spec.direction()!r}")

        return genanki.Note(
            model=YANKI_CARD_MODEL,
            fields=[
                self.text_field().render_anki(),
                self.more_field().render_anki(),
                self.media_field().render_anki(),
                text_to_media,
                media_to_text,
            ],
            guid=genanki.guid_for(self.note_id),
            tags=self.spec.config.tags,
        )

    def to_dict(self, base_url=""):
        """Recursively convert to dict."""
        return {
            "deck_id": self.deck_id,
            "note_id": self.note_id,
            "text_html": self.text_field().render_html(base_url=base_url),
            "media_html": self.media_field().render_html(base_url=base_url),
            "more_html": self.more_field().render_html(base_url=base_url),
            "media_paths": sorted(self.media_paths()),
            "direction": self.spec.direction(),
            "tags": sorted(self.spec.config.tags),
            "video_url": self.spec.video_url(),
            "source_path": self.spec.source_path,
            "line_number": self.spec.line_number,
        }


@dataclass(frozen=True)
class FinalDeck:
    deck_id: int
    title: str
    source_path: str
    spec: DeckSpec
    notes_by_id: dict

    def id(self):
        return self.deck_id

    def notes(self):
        """Get notes in the same order as the .deck file."""
        return sorted(
            self.notes_by_id.values(),
            key=lambda n: n.spec.line_number,
        )

    def media_paths(self):
        """Get media paths used in this deck."""
        for note in self.notes_by_id.values():
            yield from note.media_paths()

    def save_to_package(self, package):
        deck = genanki.Deck(self.deck_id, self.title)
        LOGGER.debug(f"New deck [{self.deck_id}]: {self.title}")

        for note in self.notes():
            deck.add_note(note.genanki_note())
            LOGGER.debug(
                f"Added note {note.note_id!r}: {note.content_fields()}"
            )

            for media_path in note.media_paths():
                package.media_files.append(media_path)
                LOGGER.debug(
                    f"Added media file for {note.note_id!r}: {media_path!r}"
                )

        package.decks.append(deck)

    def save_to_file(self, path=None):
        if path:
            path = Path(path)
        else:
            path = Path(self.source_path).with_suffix(".apkg")

        package = genanki.Package([])
        self.save_to_package(package)
        package.write_to_file(path)
        LOGGER.info(f"Wrote deck {self.title} to file {path}")

        return path

    def to_dict(self, base_url=""):
        """Recursively convert to dict."""
        return {
            "deck_id": self.deck_id,
            "title": self.title,
            "source_path": self.source_path,
            "notes": [note.to_dict(base_url=base_url) for note in self.notes()],
        }


class Deck:
    def __init__(
        self,
        spec: DeckSpec,
        video_options: VideoOptions,
    ):
        self.spec = spec
        self.video_options = video_options
        self.notes_by_id = {}
        for note_spec in spec.note_specs:
            self.add_note(Note(note_spec, video_options=video_options))

    async def finalize_async(self):
        async def finalize_note_async(collection, note, deck_id):
            final_note = await note.finalize_async(deck_id)
            collection[final_note.note_id] = final_note

        final_notes = {}
        async with asyncio.TaskGroup() as group:
            for note in self.notes():
                group.create_task(
                    finalize_note_async(final_notes, note, self.id())
                )

        return FinalDeck(
            deck_id=self.id(),
            title=self.title(),
            source_path=self.source_path(),
            spec=self.spec,
            notes_by_id=final_notes,
        )

    def id(self):
        return name_to_id(self.title())

    def title(self):
        return self.spec.title

    def source_path(self):
        return self.spec.source_path

    def notes(self):
        """Get notes in the same order as the .deck file."""
        return sorted(
            self.notes_by_id.values(),
            key=lambda n: n.spec.line_number,
        )

    def add_note(self, note):
        id = note.note_id()
        if id in self.notes_by_id:
            note.spec.error(f"Note with id {id!r} already exists in deck")
        self.notes_by_id[id] = note



from .config import (
    NOTE_VARIABLES,
    NoteConfigFrozen,
    find_invalid_format,
)
from .model import DeckSpec, NoteSpec
from .parser import DeckFilesParser

__all__ = [
    "NOTE_VARIABLES",
    "DeckFilesParser",
    "DeckSpec",
    "NoteConfigFrozen",
    "NoteSpec",
    "find_invalid_format",
]



import io

from yanki.cli.decks import DeckSource
from yanki.video import VideoOptions

REFERENCE_DECK = """
title: a
tags: +abc
file:///a A
tags: +bcd
file:///b B
    tags: +b
file:///c C
tags: -abc +def
file:///d D
tags: -bcd
file:///e E
    tags: +e
file:///f F
tags: -def
"""


def filter_deck_tags(include=frozenset(), exclude=frozenset()):
    input = io.StringIO(REFERENCE_DECK)
    input.name = "-"

    source = DeckSource(
        files=[input], tags_include=set(include), tags_exclude=set(exclude)
    )

    return "".join(
        [
            spec.text()
            for deck in source.read_specs()
            for spec in deck.note_specs
        ]
    )


def test_filters():
    assert filter_deck_tags() == "ABCDEF"
    assert filter_deck_tags(include=["abc"]) == "ABC"
    assert filter_deck_tags(exclude=["abc"]) == "DEF"
    assert filter_deck_tags(include=["abc"], exclude=["abc"]) == ""
    assert filter_deck_tags(include=["abc"], exclude=["def"]) == "ABC"
    assert filter_deck_tags(include=["abc"], exclude=["bcd"]) == "A"


def test_multiple_include():
    assert filter_deck_tags(include=["abc", "bcd"]) == "BC"
    assert filter_deck_tags(include=["abc", "bcd"], exclude=["b"]) == "C"


def test_multiple_exclude():
    assert filter_deck_tags(exclude=["abc", "bcd"]) == "EF"
    assert filter_deck_tags(include=["e"], exclude=["abc", "bcd"]) == "E"


def test_read_decks_sorted(deck_1_path, deck_2_path, cache_path):
    decks = DeckSource(
        files=[
            deck_2_path.open("r", encoding="utf_8"),
            deck_1_path.open("r", encoding="utf_8"),
        ]
    ).read_sorted(VideoOptions(cache_path))

    assert len(decks) == 2
    assert decks[0].title() == "Test::Reference deck"
    assert decks[1].title() == "Test::Reference deck::2"


def test_read_final_decks_sorted(deck_1_path, deck_2_path, cache_path):
    decks = DeckSource(
        files=[
            deck_2_path.open("r", encoding="utf_8"),
            deck_1_path.open("r", encoding="utf_8"),
        ]
    ).read_final_sorted(VideoOptions(cache_path))

    assert len(decks) == 2
    assert decks[0].title == "Test::Reference deck"
    assert decks[1].title == "Test::Reference deck::2"



import functools
import html
import re
from pathlib import Path
from urllib.parse import quote

import docutils.core
import mistletoe

# Regular expression to find http:// URLs in text.
URL_REGEX = re.compile(
    r"""
    # URL with no surrounding parentheses
    (?<!\() \b(https?://[.?!,;:a-z0-9$_+*\'()/&=@#-]*[a-z0-9$_+*\'()/&=@#-])
    # URL with surrounding parentheses
    | (?<=\() (https?://[.?!,;:a-z0-9$_+*\'()/&=@#-]*[a-z0-9$_+*\'()/&=@#-])
        (?=\))
    # URL with an initial parenthesis
    | (?<=\() (https?://[.?!,;:a-z0-9$_+*\'()/&=@#-]*[a-z0-9$_+*\'()/&=@#-])
        (?!\))
    """,
    flags=re.IGNORECASE | re.VERBOSE,
)


def rst_to_html(rst):
    # From https://wiki.python.org/moin/reStructuredText#The_.22Cool.22_Way
    parts = docutils.core.publish_parts(source=rst, writer_name="html5")
    return parts["body_pre_docinfo"] + parts["fragment"]


@functools.cache
def raw_to_html(raw):
    if raw.startswith("md:"):
        return mistletoe.markdown(raw[3:])
    if raw.startswith("rst:"):
        return rst_to_html(raw[4:])
    if raw.startswith("html:"):
        return raw[5:]
    return (
        URL_REGEX.sub(r'<a href="\1">\1</a>', html.escape(raw))
        .rstrip()
        .replace("\n", "<br/>")
    )


class Fragment:
    def __init__(self, raw):
        self.raw = raw

    def media(self):
        return []

    def media_paths(self):
        return []

    def render_anki(self):
        return self.render_html("")

    def render_html(self, _base_url=""):
        return raw_to_html(self.raw)

    def __str__(self):
        return self.render_anki()

    def __repr__(self):
        return repr(self.render_anki())


class MediaFragment(Fragment):
    def __init__(self, path: Path, media):
        self.path = path
        self._media = media

    def file_name(self):
        return self.path.name

    def path_in_base(self, _base_url):
        return

    def anki_filename(self):
        """Get the filename encoded for Anki."""
        # FIXME need to prevent characters that break Anki.
        # html.escape() breaks Anki. A literal single quote (') is escaped as
        # &#x27;, which then gets transformed to &amp;%23x27; at some point.
        return self.file_name()

    def html_path_in_base(self, base_url: str):
        """Get the path relative to base_url, and encoded for HTML.

        base_url may be a absolute or relative URL.
        """
        quoted_file_name = quote(self.file_name(), encoding="utf_8")
        if base_url == "":
            return html.escape(quoted_file_name)
        return html.escape(f"{base_url}/{quoted_file_name}")

    def media(self):
        return [self._media]

    def media_paths(self):
        return [str(self.path)]


class ImageFragment(MediaFragment):
    def render_anki(self):
        return f'<img src="{self.anki_filename()}" />'

    def render_html(self, base_url=""):
        return f'<img src="{self.html_path_in_base(base_url)}" />'


class VideoFragment(MediaFragment):
    def render_anki(self):
        return f"[sound:{self.anki_filename()}]"

    def render_html(self, base_url="."):
        return (
            "<video controls playsinline "
            f'src="{self.html_path_in_base(base_url)}"></video>'
        )


class Field:
    def __init__(self, fragments: list[Fragment] | None = None):
        if fragments is None:
            fragments = []
        self.fragments = fragments

    def add_fragment(self, fragment: Fragment):
        self.fragments.append(fragment)

    def media(self):
        for fragment in self.fragments:
            yield from fragment.media()

    def media_paths(self):
        for fragment in self.fragments:
            yield from fragment.media_paths()

    def render_anki(self):
        return "".join([fragment.render_anki() for fragment in self.fragments])

    def render_html(self, base_url=""):
        return "".join(
            [fragment.render_html(base_url) for fragment in self.fragments]
        )

    def __str__(self):
        return self.render_anki()

    def __repr__(self):
        return repr(list(self.fragments))



import asyncio
import functools
import hashlib
import json
import logging
import math
import re
import shlex
from dataclasses import dataclass
from multiprocessing import cpu_count
from pathlib import Path
from urllib.parse import parse_qs, urlparse

import ffmpeg
import yt_dlp

from yanki.errors import ExpectedError
from yanki.utils import (
    NotFileURLError,
    atomic_open,
    chars_in,
    file_not_empty,
    file_url_to_path,
    get_key_path,
)

LOGGER = logging.getLogger(__name__)

STILL_FORMATS = frozenset(["png", "jpeg", "jpg"])
FILENAME_ILLEGAL_CHARS = '/"[]:'
MORE_INFO_VERSION = 2

# For parsing ffmpeg output.
CROPDETECT_RE = re.compile(
    rb"\[Parsed_cropdetect_.* t:(\d+\.\d+) .* crop=(\S+)"
)


class BadURLError(ExpectedError):
    pass


class FFmpegError(RuntimeError):
    def __init__(
        self,
        command="ffmpeg",
        command_line=None,
        stdout=None,
        stderr=None,
        exit_code=None,
    ):
        super().__init__(f"Error running {command}")
        self.command = command
        if command_line:
            self.add_note(f"Command run: {shlex.join(command_line)}")
        self.command_line = command_line
        self.stdout = stdout
        self.stderr = stderr
        self.exit_code = exit_code


@dataclass
class VideoOptions:
    """Options for processing videos."""

    cache_path: Path
    progress: bool = False
    reprocess: bool = False
    concurrency: int = cpu_count()

    @functools.cached_property
    def semaphore(self):
        return asyncio.Semaphore(self.concurrency)


# Example YouTube video URLs:
# https://gist.github.com/rodrigoborgesdeoliveira/987683cfbfcc8d800192da1e73adc486
#
#   https://www.youtube.com/watch?v=n1PjPqcHswk
#   https://youtube.com/watch/lalOy8Mbfdc
def youtube_url_to_id(url_str, url, query):
    """Get YouTube video ID, e.g. lalOy8Mbfdc, from a youtube.com URL."""
    if len(query.get("v", [])) == 1:
        return query["v"][0]

    try:
        path = url.path.split("/")
        if path[0] == "" and path[1] in {"watch", "v"}:
            return path[2]
    except IndexError:
        # Fall through to error.
        pass

    raise BadURLError(f"Unknown YouTube URL format: {url_str}")


# URLs like http://youtu.be/lalOy8Mbfdc
def youtu_be_url_to_id(url_str, url, _query):
    """Get YouTube video ID, e.g. lalOy8Mbfdc, from a youtu.be URL."""
    try:
        path = url.path.split("/")
        if path[0] == "":
            return path[1].split("&")[0]
    except IndexError:
        # Fall through to error.
        pass

    raise BadURLError(f"Unknown YouTube URL format: {url_str}")


def url_to_id(url_str):
    """Turn video URL into an ID string that can be part of a file name."""
    url = urlparse(url_str)
    query = parse_qs(url.query)

    try:
        domain = "." + url.netloc.lower()
        if domain.endswith(".youtube.com"):
            return "youtube=" + youtube_url_to_id(url_str, url, query)
        if domain.endswith(".youtu.be"):
            return "youtube=" + youtu_be_url_to_id(url_str, url, query)
    except BadURLError:
        # Try to load the URL with yt_dlp and see what happens.
        pass

    # FIXME check this against FILENAME_ILLEGAL_CHARS somehow
    return (
        url_str.replace("\\", "\\\\")
        .replace("|", r"\|")
        .replace('"', r"\'")
        .replace("[", r"\(")
        .replace("]", r"\)")
        .replace(":", r"\=")
        .replace("/", "|")
    )


class Video:
    def __init__(
        self,
        url,
        options,
        working_dir=Path(),
        logger=LOGGER,
    ):
        self.url = url
        self.working_dir = working_dir
        self.options = options
        self.logger = logger

        # self.options is read only, and this will be set to false after
        # reprocessing so we don’t do it over and over.
        self.reprocess = options.reprocess

        self.id = url_to_id(url)
        if invalid := chars_in(FILENAME_ILLEGAL_CHARS, self.id):
            invalid = "".join(invalid)
            raise BadURLError(
                f"Invalid characters ({invalid}) in video ID: {self.id!r}"
            )

        self._raw_metadata = None
        self._format = None
        self._strip_audio = False
        self._strip_video = False
        self._clip = None
        self._crop = None
        self._overlay_text = ""
        self._slow = None

        # Only available after finalizing (or calling their generation methods):
        self._cached_more_info = None
        self._cached_parameters = None

    def cached(self, filename):
        return self.options.cache_path / filename

    def info_cache_path(self):
        return self.cached(f"info_{self.id}.json")

    def more_info_cache_path(self):
        return self.cached(
            f"more_info_{self.id}_clip={self.file_safe_clip()}.json"
        )

    def raw_video_cache_path(self):
        return self.cached("raw_" + self.id + "." + self.info()["ext"])

    def raw_metadata_cache_path(self):
        return self.cached(f"ffprobe_raw_{self.id}.json")

    async def processed_video_cache_path_async(self, prefix="processed_"):
        parameters = "_".join(await self.parameters_list_async())

        if len(parameters) > 60 or chars_in(FILENAME_ILLEGAL_CHARS, parameters):
            parameters = hashlib.blake2b(
                parameters.encode(encoding="utf_8"),
                digest_size=16,
                usedforsecurity=False,
            ).hexdigest()
        return self.cached(
            f"{prefix}{self.id}_{parameters}.{self.output_ext()}"
        )

    def _download_info(self):
        try:
            path = file_url_to_path(self.url)
            return {
                "title": path.stem,
                "ext": path.suffix[1:],
            }
        except NotFileURLError:
            pass

        try:
            with self._yt_dlp() as ydl:
                self.logger.info(f"getting info about {self.url!r}")
                return ydl.sanitize_info(
                    ydl.extract_info(self.url, download=False)
                )
        except yt_dlp.utils.YoutubeDLError as error:
            # This is an ExpectedError, so the __cause__ won’t normally be
            # displayed, so it’s included in the message.
            raise BadURLError(
                f"Error downloading {self.url!r}: {error}"
            ) from error

    @functools.cache
    def info(self):
        try:
            with self.info_cache_path().open("r", encoding="utf_8") as file:
                return json.load(file)
        except FileNotFoundError:
            # Either the file wasn’t found or wasn’t valid JSON. We use `pass`
            # to avoid adding this exception to the context of new exceptions.
            pass

        info = self._download_info()
        with atomic_open(self.info_cache_path()) as file:
            json.dump(info, file)
        return info

    def title(self):
        return self.info()["title"]

    def refresh_raw_metadata(self):
        self.logger.debug(f"refresh raw metadata: {self.raw_video()}")
        try:
            self._raw_metadata = ffmpeg.probe(self.raw_video())
        except ffmpeg.Error as error:
            raise FFmpegError(
                command="ffprobe", stdout=error.stdout, stderr=error.stderr
            ) from error

        with atomic_open(self.raw_metadata_cache_path()) as file:
            json.dump(self._raw_metadata, file)

        return self._raw_metadata

    # This will refresh metadata once if it doesn’t find the passed path the
    # first time.
    def raw_metadata(self, *key_path):
        try:
            # FIXME? Track if ffprobe was already run and don’t run it again.
            if self._raw_metadata:
                return get_key_path(self._raw_metadata, key_path)

            metadata_cache_path = self.raw_metadata_cache_path()
            if (
                metadata_cache_path.stat().st_mtime
                >= self.raw_video().stat().st_mtime
            ):
                # Metadata isn’t older than raw video.
                with metadata_cache_path.open("r", encoding="utf_8") as file:
                    self._raw_metadata = json.load(file)
                    return get_key_path(self._raw_metadata, key_path)
        except (FileNotFoundError, json.JSONDecodeError, KeyError, IndexError):
            # Either the file wasn’t found, wasn’t valid JSON, or it didn’t have
            # the key path. We use `pass` here to avoid adding this exception to
            # the context of new exceptions.
            pass

        return get_key_path(self.refresh_raw_metadata(), key_path)

    def get_fps(self):
        for stream in self.raw_metadata("streams"):
            if stream["codec_type"] == "video":
                division = stream["avg_frame_rate"].split("/")
                if len(division) == 0:
                    continue

                fps = float(division.pop(0))
                for divisor in division:
                    fps /= float(divisor)

                return fps

        raise BadURLError(f"Could not get FPS for media URL {self.url!r}")

    # Expects spec without whitespace
    def time_to_seconds(self, spec, on_none=None):
        """Convert a time spec like 1:01.02 or 4F to decimal seconds."""
        if spec == "" or spec is None:
            return on_none

        if isinstance(spec, (float, int)):
            return float(spec)

        if spec[-1] in "Ff":
            # Frame number
            return int(spec[:-1]) / self.get_fps()
        if spec[-1] in "Ss":
            # Second (s), millisecond (ms), or microsecond (us) suffix
            if spec[-2] in "Mm":
                return float(spec[:-2]) / 1_000
            if spec[-2] in "Uuµ":
                return float(spec[:-2]) / 1_000_000
            return float(spec[:-1])

        # [-][HH]:[MM]:[SS.mmm...]
        sign = 1
        if spec.startswith("-"):
            spec = spec[1:]
            sign = -1

        # FIXME? this acccepts 3.3:500:67.8:0:1.2
        sum = 0
        for part in spec.split(":"):
            sum = sum * 60 + float(part)

        return sign * sum

    def cropdetect(self):
        """Detect black borders to crop.

        more_info_async() must be called first.
        """
        more = self.more_info()

        try:
            # FIXME figure out largest crop
            return more["cropdetect"][0][1]
        except (IndexError, KeyError, TypeError):
            return None

    async def load_more_info_async(self):
        """Load more information about the contents of the media.

        This returns a `dict` with keys:
          * `version`: the current version of the more_info algorithm so that
            old data can be invalidated.
          * `cropdetect`:
            * `[(time, "crop"), ...]`, e.g. `[(0.067, "1920:1072:0:4"), ...]`
            * `None`: the video stream was stripped

        https://ayosec.github.io/ffmpeg-filters-docs/7.0/Filters/Video/cropdetect.html
        """
        if not self.wants_video():
            self._cached_more_info = {
                "version": MORE_INFO_VERSION,
                "cropdetect": None,
            }
            return self._cached_more_info

        # Only process the part of the media we care about.
        in_options = self.clip_to_ffmpeg_input_options(self._clip)
        out_options = self.clip_to_ffmpeg_output_options(self._clip)

        video = ffmpeg.input(str(self.raw_video()), **in_options)["v"]
        video = video.filter("cropdetect", round=2)

        (_, err) = await self.run_async(
            video.output("-", format="null", **out_options)
        )
        # FIXME use metadata=mode=print?

        # (time, "crop"), e.g. (2.369, "1920:1072:0:4")
        cropdetect = [
            (float(matches[1]), matches[2].decode("utf_8"))
            for line in err.split(b"\n")
            if (matches := CROPDETECT_RE.search(line))
        ]

        self._cached_more_info = {
            "version": MORE_INFO_VERSION,
            "cropdetect": cropdetect,
        }
        return self._cached_more_info

    async def more_info_async(self):
        if self._cached_more_info:
            return self._cached_more_info

        path = self.more_info_cache_path()
        try:
            with path.open("r", encoding="utf_8") as file:
                self._cached_more_info = json.load(file)
                if self._cached_more_info is not None:
                    # No version is equivalent to version 1.
                    version = self._cached_more_info.get("version", 1)
                    if version == MORE_INFO_VERSION:
                        return self._cached_more_info
                    self.logger.info(
                        f"Discarding more_info with bad version {version!r} "
                        f"(expected {MORE_INFO_VERSION!r}) at {path}"
                    )
                    self._cached_more_info = None
        except FileNotFoundError:
            # Either the file wasn’t found or wasn’t valid JSON. We use `pass`
            # to avoid adding this exception to the context of new exceptions.
            pass

        await self.load_more_info_async()
        with atomic_open(path) as file:
            json.dump(self._cached_more_info, file)
        return self._cached_more_info

    def more_info(self):
        """Get extra information from finalized video."""
        if self._cached_more_info is None:
            raise ValueError("more_info() called on un-finalized Video")
        return self._cached_more_info

    async def finalize_async(self):
        # Ensure that the video is fully processed and that everything can be
        # accessed synchronously.
        await self.more_info_async()
        await self.processed_video_async()
        return self

    def clip(self, start_spec, end_spec):
        start = self.time_to_seconds(start_spec, on_none=0)
        end = self.time_to_seconds(end_spec, on_none=None)
        if end is not None and end - start <= 0:
            raise ValueError(
                "Cannot clip video to 0 or fewer seconds "
                f"({start_spec!r} to {end_spec!r})"
            )
        self._clip = (start, end)

    def snapshot(self, time_spec):
        self._clip = self.time_to_seconds(time_spec, on_none=None)

    def crop(self, crop):
        if crop in {"none", ""}:
            crop = None
        self._crop = crop

    def overlay_text(self, text):
        self._overlay_text = text

    def audio(self, audio):
        self._strip_audio = audio == "strip"

    def video(self, video):
        self._strip_video = video == "strip"

    def slow(self, start=0, end=None, amount=2):
        """Slow (or speed up) part of the video."""
        start = self.time_to_seconds(start, on_none=0)
        end = self.time_to_seconds(end, on_none=None)

        if (end is not None and end == start) or amount == 1:
            # Nothing is affected
            self._slow = None
        else:
            self._slow = (start, end, float(amount))

    def format(self, extension: str | None):
        if extension is None:
            self._format = None
        else:
            self._format = extension.lower()

    def output_ext(self):
        if self._format is not None:
            return self._format
        if self.is_still():
            return "jpeg"
        return "mp4"

    def is_still(self):
        return (
            isinstance(self._clip, float)
            or self._format in STILL_FORMATS
            or "duration" not in self.raw_metadata("format")
        )

    def has_audio(self):
        """If the raw video contains an audio stream."""
        for stream in self.raw_metadata("streams"):
            if stream["codec_type"] == "audio":
                return True
        return False

    def wants_audio(self):
        """If the output should include an audio stream."""
        return (
            not self._strip_audio and self.has_audio() and not self.is_still()
        )

    def has_video(self):
        """If the raw video contains a video stream or image."""
        for stream in self.raw_metadata("streams"):
            if stream["codec_type"] == "video":
                return True
        return False

    def wants_video(self):
        """If the output should include a video stream or image."""
        return not self._strip_video and self.has_video()

    @functools.cache
    def raw_video(self):
        try:
            # If it’s a file:// URL, then there’s no need to cache.
            source_path = self.working_dir / file_url_to_path(self.url)
        except NotFileURLError:
            pass
        else:
            self.logger.info(f"using local raw video {source_path}")
            return source_path

        if "ext" not in self.info():
            raise BadURLError(f"Invalid media URL {self.url!r}")

        path = self.raw_video_cache_path()
        if path.exists() and path.stat().st_size > 0:
            # Already cached, and we can’t check if it’s out of date.
            return path

        self.logger.info(f"downloading raw video to {path}")

        with self._yt_dlp(outtmpl={"default": str(path)}) as ydl:
            # FIXME why not use the in-memory info?
            if error := ydl.download_with_info_file(self.info_cache_path()):
                # FIXME??!
                raise RuntimeError(error)

        return path

    def clip_to_ffmpeg_input_options(self, clip):
        """Input options for ffmpeg based on real clip.

        Used by load_more_info_async() and processed_video_async().
        """
        options = {}
        match clip:
            case None:
                pass
            case float(snapshot_time):
                options["ss"] = snapshot_time
            case (start, end):
                if end is not None:
                    options["t"] = end - start
                if start:
                    options["ss"] = start
            case other:
                raise ValueError(f"parameter may not be {other!r}")

        return options

    def clip_to_ffmpeg_output_options(self, clip):
        """Output options for ffmpeg based on real clip.

        Used by load_more_info_async() and processed_video_async().
        """
        match clip:
            case None | (_, _):
                return {}
            case float(_snapshot_time):
                return {
                    "frames:v": "1",
                    "q:v": "2",  # JPEG quality
                }
            case other:
                raise ValueError(f"parameter may not be {other!r}")

    def file_safe_clip(self):
        """Get the clip in a filesystem safe format. Does not calculate auto."""
        match self._clip:
            case None | (0, None):
                return "none"
            case float(snapshot_time):
                return snapshot_time
            case (float(_) | int(0) as start, None | float(_) as end):
                return f"({start},{end})"  # No space
            case other:
                raise ValueError(f"Unexpected _clip: {other!r}")

    async def actual_crop_async(self):
        if self._crop == "auto":
            await self.more_info_async()
            return self.cropdetect()
        return self._crop

    async def actual_clip_async(self):
        match self._clip:
            case None:
                return None
            case float(snapshot_time):
                return snapshot_time
            case (float(_) | int(0) as start, None | float(_) as end):
                return (start, end)
            case other:
                raise ValueError(f"Unexpected _clip: {other!r}")

    async def parameters_async(self):
        """Get parameters for producing the video as a dict."""
        self._cached_parameters = {}

        if self._strip_audio:
            self._cached_parameters["audio"] = "strip"
        if self._strip_video:
            self._cached_parameters["video"] = "strip"
        if self._crop is not None:
            self._cached_parameters["crop"] = await self.actual_crop_async()
        if self._overlay_text != "":
            self._cached_parameters["overlay_text"] = self._overlay_text
        if self._slow is not None:
            self._cached_parameters["slow"] = self._slow

        match await self.actual_clip_async():
            case float(time):
                self._cached_parameters["snapshot"] = time
            case (start, end):
                self._cached_parameters["clip"] = (start, end)

        return self._cached_parameters

    # FIXME seems like there should be a separate FinalVideo class with this.
    def parameters_list(self):
        """Get video parameters list[str] (finalized version)."""
        if self._cached_parameters is None:
            raise ValueError("parameters_list() called on un-finalized Video")
        return sorted(
            [
                f"{key}={value!r}"
                for key, value in self._cached_parameters.items()
            ]
        )

    async def parameters_list_async(self):
        """Get parameters for producing the video as list[str]."""
        return sorted(
            [
                f"{key}={value!r}"
                for key, value in (await self.parameters_async()).items()
            ]
        )

    async def processed_video_async(self):
        output_path = await self.processed_video_cache_path_async()
        if not self.reprocess and file_not_empty(output_path):
            return output_path

        # Only reprocess once per run.
        self.reprocess = False

        parameters = " ".join(await self.parameters_list_async())
        self.logger.info(f"processing with ({parameters}) to {output_path}")

        clip = await self.actual_clip_async()
        in_options = self.clip_to_ffmpeg_input_options(clip)
        out_options = self.clip_to_ffmpeg_output_options(clip)

        stream = ffmpeg.input(str(self.raw_video()), **in_options)
        output_streams = {}

        if self.wants_video():
            # Video stream is not being stripped
            video = stream["v"]
            if crop := await self.actual_crop_async():
                # FIXME kludge; doesn’t handle named params
                video = video.filter("crop", *crop.split(":"))

            video = video.filter("scale", -2, 500)

            if self._overlay_text:
                video = video.drawtext(
                    text=self._overlay_text,
                    x=20,
                    y=20,
                    font="Arial",
                    fontcolor="white",
                    fontsize=48,
                    box=1,
                    boxcolor="black@0.5",
                    boxborderw=20,
                )

            output_streams["v"] = video

        if self.wants_audio():
            # Audio stream is not being stripped
            audio = stream["a"]
            output_streams["a"] = audio

        output_streams = self._try_apply_slow(output_streams)
        if isinstance(output_streams, dict):
            output_streams = output_streams.values()
        else:
            output_streams = [output_streams]

        with atomic_open(output_path, encoding=None) as file:
            file.close()
            stream = ffmpeg.output(
                *output_streams, file.name, **out_options
            ).overwrite_output()

            await self.run_async(stream)

        return output_path

    async def run_async(self, stream):
        command = stream.compile()
        self.logger.debug(f"Run {shlex.join(command)}")

        async with self.options.semaphore:
            process = await asyncio.create_subprocess_exec(
                *command,
                stdin=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.PIPE,
            )

            stdout, stderr = await process.communicate()

        if process.returncode:
            raise FFmpegError(
                command_line=command,
                stderr=stderr,
                exit_code=process.returncode,
            )
        return stdout, stderr

    # Expect { 'v': video?, 'a' : audio? } depending on if -vn and -an are set.
    def _try_apply_slow(self, streams):  # noqa: C901 PLR0912 (FIXME complex)
        if self._slow is None:
            return streams

        # These are already floats (or None for end):
        (start, end, amount) = self._slow

        wants_video = self.wants_video()
        wants_audio = self.wants_audio()
        parts = []
        i = 0

        if wants_video:
            vsplit = streams["v"].split()
        if wants_audio:
            asplit = streams["a"].asplit()

        if start != 0:
            if wants_video:
                parts.append(
                    vsplit[i]
                    .filter("trim", start=0, end=start)
                    .filter("setpts", "PTS-STARTPTS")
                )
            if wants_audio:
                parts.append(
                    asplit[i]
                    .filter("atrim", start=0, end=start)
                    .filter("asetpts", "PTS-STARTPTS")
                )
            i += 1

        if end is None:
            expression = {"start": start}
        else:
            expression = {"start": start, "end": end}

        if wants_video:
            parts.append(
                vsplit[i]
                .filter("trim", **expression)
                .filter("setpts", "PTS-STARTPTS")
                .setpts(f"{amount}*PTS")
            )
        if wants_audio:
            part = (
                asplit[i]
                .filter("atrim", **expression)
                .filter("asetpts", "PTS-STARTPTS")
            )

            if amount < 0.01:
                # FIXME validate on parse
                raise ValueError("Cannot slow audio by less than 0.01")
            if amount > 2:
                twos_count = math.floor(math.log2(amount))
                for _ in range(twos_count):
                    part = part.filter("atempo", 0.5)
                last_amount = amount / 2**twos_count
                if last_amount != 1:
                    part = part.filter("atempo", 1 / last_amount)
            else:
                part = part.filter("atempo", 1 / amount)

            parts.append(part)
        i += 1

        if end is not None:
            if wants_video:
                parts.append(
                    vsplit[i]
                    .filter("trim", start=end)
                    .filter("setpts", "PTS-STARTPTS")
                )
            if wants_audio:
                parts.append(
                    asplit[i]
                    .filter("atrim", start=end)
                    .filter("asetpts", "PTS-STARTPTS")
                )

        return ffmpeg.concat(*parts, v=int(wants_video), a=int(wants_audio))

    def _yt_dlp(self, **kwargs):
        """Run yt_dlp."""
        return yt_dlp.YoutubeDL(
            {
                "logtostderr": True,
                "noprogress": not self.options.progress,
                "skip_unavailable_fragments": False,
                "quiet": True,
                **kwargs,
            }
        )
